import unittest
from unittest import skip
from cryptos import coins_async
from cryptos.testing.testcases_async import BaseAsyncCoinTestCase
from typing import List


class TestDashTestnet(BaseAsyncCoinTestCase):
    name = "Dash Testnet"
    coin = coins_async.Dash
    addresses = ["yh1u3ZD4kGKttnwaNXpyP85FH3eD8E99vP",
                 "yTXgT2aA32Y35VQ6N9KpFqKJKKdbidgKeZ",
                 "ySPomQ35mpKiV89LDdAM3URFSibNiXEC4J"]
    multisig_addresses: List[str] = ["8hwYpDfDw526ppvaRveUv16nG2zXwp1Z7X", "8fdMUYERYqzfxjTwEYUcAFWSicLRBA7rxn"]
    privkeys = ["098ddf01ebb71ead01fc52cb4ad1f5cafffb5f2d052dd233b3cad18e255e1db1",
                "0861e1bb62504f5e9f03b59308005a6f2c12c34df108c6f7c52e5e712a08e91401",
                "c396c62dfdc529645b822dc4eaa7b9ddc97dd8424de09ca19decce61e6732f71"]  #Private keys for above address_derivations in same order
    privkey_standard_wifs: List[str] = ["91f8DFTsmhtawuLjR8CiHNkgZGPkUqfJ45LxmENPf3k6fuX1m4N",
                                        "cMrziExc6iMV8vvAML8QX9hGDP8zNhcsKbdS9BqrRa1b4mhKvK6f",
                                        "9354Dkk67pJCfmRfMedJPhGPfZCXv2uWd9ZoVNMUtDxjUBbCVZK"]
    fee: int = 5000
    max_fee: int = 100000
    testnet: bool = True
    min_latest_height = 830385
    unspent_addresses = ["ye9FSaGnHH5A2cjJ9s2y9XTgyJZefB5huz"]
    unspent = [{'address': 'ye9FSaGnHH5A2cjJ9s2y9XTgyJZefB5huz',
                'height': 830413,
                'tx_hash': 'a9f58c5bf785528edcda59a1e1f679a690be2f2c036e194bb2bc3f6ed0b5a9a5',
                'tx_pos': 0,
                'value': 378260000}]
    unspents = unspent
    balance = {'confirmed': 378260000, 'unconfirmed': 0}
    balances = [{'address': unspent_addresses[0]} | dict(balance)]
    history = [
        {'height': 830413,
         'tx_hash': 'a9f58c5bf785528edcda59a1e1f679a690be2f2c036e194bb2bc3f6ed0b5a9a5'}
    ]
    histories = [{'address': unspent_addresses[0]} | dict(history[0])]
    txid = 'a9f58c5bf785528edcda59a1e1f679a690be2f2c036e194bb2bc3f6ed0b5a9a5'
    txheight = 45550
    block_hash: str = "5a353b1bc4974919fc49c0b15f6c2226e27dfdef0fe1809cd90f67eb50e6b478"
    raw_tx: str = "020000000112f461b7fbc2607586659722ec82d81e568ea5e66308e2ea76b0b284fabefb1f010000006a47304402201fcd36ce719006af4278f0624c922df5e275be767a468fdf66c72a0de8b6bb8202206cb4fc48f0b91be14b1135856fd37c0ba9c6b670199131ea3d8150979f652322012103c54188322b2e2639a465f63c9a30ee620c3bd3af157e1c5781d34d2f779dca9ffeffffff0220ca8b16000000001976a914c384950342cb6f8df55175b48586838b03130fad88ac1c181c50000000001976a9146fd4df7d158c66f9f00c1247ad64a9e7d6a6c33388acc5ab0c00"
    expected_tx_verbose_keys: List[str] = ['blockhash', 'blocktime', 'chainlock', 'confirmations', 'hex', 'instantlock', 'instantlock_internal', 'locktime', 'size', 'time', 'txid', 'type', 'version', 'vin', 'vout']

    def test_standard_wif_ok(self):
        self.assertStandardWifOK()

    async def test_balance(self):
        await self.assertBalanceOK()

    async def test_balances(self):
        await self.assertBalancesOK()

    async def test_unspent(self):
        await self.assertUnspentOK()

    async def test_unspents(self):
        await self.assertUnspentsOK()

    async def test_merkle_proof(self):
        await self.assertMerkleProofOK()

    async def test_history(self):
        await self.assertHistoryOK()

    async def test_histories(self):
        await self.assertHistoriesOK()

    async def test_balance_merkle_proven(self):
        await self.assertBalanceMerkleProvenOK()

    async def test_balances_merkle_proven(self):
        await self.assertBalancesMerkleProvenOK()

    async def test_block_header(self):
        await self.assertBlockHeaderOK()

    async def test_block_headers(self):
        await self.assertBlockHeadersOK()

    async def test_gettx(self):
        await self.assertGetTXOK()

    async def test_getverbosetx(self):
        await self.assertGetVerboseTXOK()

    async def test_gettxs(self):
        await self.assertGetTXOK()

    async def test_transaction(self):
        """
        Sample transaction:
        TxID: 23ebe2519c10803ff4d3cab0013d80b63e2ba90103c98084aacbc79c78dd736f
        """
        await self.assertTransactionOK()

    @unittest.skip("Intermittent failure")
    async def test_transaction_multisig(self):
        """
        Sample transaction:
        TxID:
        """
        await self.assertMultiSigTransactionOK()

    @unittest.skip("Intermittent failure")
    async def test_sendmulti_recipient_tx(self):
        """
        Sample transaction:
        TxID: 6fb0071bd94dff8de1f9784fd30aadeebd2d8ca27f5451b7725e05c868e98593
        """
        await self.assertSendMultiRecipientsTXOK()

    async def test_send(self):
        """
        Sample transaction:
        TxID: 81a22353f6dc45ae9b32a8fdea26b74fa71050c6993959cb07ce7092b26b1287
        """
        await self.assertSendOK()

    async def test_subscribe_block_headers(self):
        await self.assertSubscribeBlockHeadersOK()

    async def test_subscribe_block_headers_sync(self):
        await self.assertSubscribeBlockHeadersSyncCallbackOK()

    async def test_latest_block(self):
        await self.assertLatestBlockOK()

    async def test_confirmations(self):
        await self.assertConfirmationsOK()

    async def test_subscribe_address(self):
        await self.assertSubscribeAddressOK()

    async def test_subscribe_address_sync(self):
        await self.assertSubscribeAddressSyncCallbackOK()

    async def test_subscribe_address_transactions(self):
        await self.assertSubscribeAddressTransactionsOK()

    async def test_subscribe_address_transactions_sync(self):
        await self.assertSubscribeAddressTransactionsSyncOK()



# Electrum - lightweight Bitcoin client
# Copyright (C) 2011 Thomas Voegtlin
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import re
from typing import AsyncGenerator, List, Any, Optional


def user_dir(appname: str) -> Optional[str]:
    if os.name == 'posix':
        return os.path.join(os.environ["HOME"], ".%s" % appname.lower())
    elif "APPDATA" in os.environ:
        return os.path.join(os.environ["APPDATA"], appname.capitalize())
    elif "LOCALAPPDATA" in os.environ:
        return os.path.join(os.environ["LOCALAPPDATA"], appname.capitalize())
    else:
        # raise Exception("No home directory found in environment variables.")
        return


async def alist(generator: AsyncGenerator[Any, None]) -> List[Any]:
    return [i async for i in generator]


def is_hex(text: str) -> Optional[re.Match[str]]:
    return re.match('^[0-9a-fA-F]*$', text)



import unittest

from cryptos import coins_async
from cryptos.testing.testcases_async import BaseAsyncCoinTestCase
from typing import List
from unittest import mock


class TestDoge(BaseAsyncCoinTestCase):
    name = "Dogecoin"
    coin = coins_async.Doge
    addresses = ['DRqYjcRNeRMWw7c3gPBv3L8i3ZJSqbm6PV',
                 'DCML95nTwBZf7p4Zfzgkv3Nm5qHqY1g17S',
                 'DBDTTTFPfyMLXSooXUXHhgUiDEFcTRnvFf']
    multisig_addresses: List[str] = ["9uxMmYZ64cQjH8LqWBeSNNv99YmMznrBuN",
                                     "9seARs8HgPPJR2tCJoUZcdKoc87FNnHZhg"]
    privkeys: List[str] = ["098ddf01ebb71ead01fc52cb4ad1f5cafffb5f2d052dd233b3cad18e255e1db1",
                           "0861e1bb62504f5e9f03b59308005a6f2c12c34df108c6f7c52e5e712a08e91401",
                           "c396c62dfdc529645b822dc4eaa7b9ddc97dd8424de09ca19decce61e6732f71"]
    privkey_standard_wifs: List[str] = ['6JCpvU7HkF7F2TGbzAvn2QBxP2WapoQNKJcF16en8yCghYr1m7W',
                                       'QNtvQAmjvG9MD1qwZCA5342p4Bs9koez8pAspPnBeoxwFxbqYcre',
                                       '6KckvyPW6MWrkKMXvhMN8ihfVKKNFzeatNq5jEdsN9RKVnJrynQ']
    fee: int = 300000
    max_fee: int = 600000
    testnet = False

    min_latest_height = 4464523
    unspent_addresses = ["DLAznsPDLDRgsVcTFWRMYMG5uH6GddDtv8"]

    balance = {'confirmed': 730696900, 'unconfirmed': 0}
    balances = [{'address': unspent_addresses[0]} | dict(balance)]

    unspent = [{'address': 'DLAznsPDLDRgsVcTFWRMYMG5uH6GddDtv8',
                 'height': 3101154,
                 'tx_hash': '770e6cb667a07c74a73ed74950224cb536ceed34dc678182a0d86c7a7703ed9c',
                 'tx_pos': 0,
                 'value': 100000000},
               {'address': 'DLAznsPDLDRgsVcTFWRMYMG5uH6GddDtv8',
                 'height': 3133347,
                 'tx_hash': '976a9e6cd24d44cce4098401398a37663973a297ab60c2f42db7dc46640ca410',
                 'tx_pos': 0,
                 'value': 69696900},
                 {'address': 'DLAznsPDLDRgsVcTFWRMYMG5uH6GddDtv8',
                  'height': 3311542,
                  'tx_hash': 'd19cd7b01a6e7320f3b5efb90dc2cc4505ac8bf32b2e6b6fae87a87615752976',
                  'tx_pos': 0,
                  'value': 420000000},
                 {'address': 'DLAznsPDLDRgsVcTFWRMYMG5uH6GddDtv8',
                  'height': 4073469,
                  'tx_hash': 'aa5cb5cccf8dadd5a42d039134b732f76226b6c09aca629cbf0489a5974578e6',
                  'tx_pos': 0,
                  'value': 141000000}]
    unspents = unspent
    history = [
        {'height': 3101154,
          'tx_hash': '770e6cb667a07c74a73ed74950224cb536ceed34dc678182a0d86c7a7703ed9c'},
        {'height': 3133347,
          'tx_hash': '976a9e6cd24d44cce4098401398a37663973a297ab60c2f42db7dc46640ca410'},
        {'height': 3311542,
          'tx_hash': 'd19cd7b01a6e7320f3b5efb90dc2cc4505ac8bf32b2e6b6fae87a87615752976'},
        {'height': 4073469,
          'tx_hash': 'aa5cb5cccf8dadd5a42d039134b732f76226b6c09aca629cbf0489a5974578e6'}]
    histories = [{'address': 'DLAznsPDLDRgsVcTFWRMYMG5uH6GddDtv8',
  'height': 3101154,
  'tx_hash': '770e6cb667a07c74a73ed74950224cb536ceed34dc678182a0d86c7a7703ed9c'},
 {'address': 'DLAznsPDLDRgsVcTFWRMYMG5uH6GddDtv8',
  'height': 3133347,
  'tx_hash': '976a9e6cd24d44cce4098401398a37663973a297ab60c2f42db7dc46640ca410'},
 {'address': 'DLAznsPDLDRgsVcTFWRMYMG5uH6GddDtv8',
  'height': 4073469,
  'tx_hash': 'aa5cb5cccf8dadd5a42d039134b732f76226b6c09aca629cbf0489a5974578e6'},
 {'address': 'DLAznsPDLDRgsVcTFWRMYMG5uH6GddDtv8',
  'height': 3311542,
  'tx_hash': 'd19cd7b01a6e7320f3b5efb90dc2cc4505ac8bf32b2e6b6fae87a87615752976'}]
    txid: str = "345c28885d265edbf8565f553f9491c511b6549d3923a1d63fe158b8000bbee2"
    txheight: int = 2046470
    block_hash: str = "404646bb8958730e133c2a363b79ef1c9c300db4134988060c6c9215e9a6694a"
    raw_tx: str = "010000000185229c0ab8804cb1e4bf465d9e3e0bcbc0a40481a5892eb03d61d7411f1fee72000000006a47304402202da8ad356685fe8f3337cd14c7a20861fc332b79a49581c062fe2f648e4b8d0b0220106fed579b94c0bf1b6bdb199c8147deb65b40f6618bab89c00262090124b70d012102960791fd0876b1b00ae11ea858d4f3f44a3184d8d4bff0336417b969b9f11809feffffff02928eccfd280000001976a91466f79e401b78766cbea8791ec76c8d6b2cf6441788ac75969b6e2b0300001976a914f594b41cfd73d290bbbb298d71e63d1c613d35ee88acf8391f00"

    def test_standard_wif_ok(self):
        self.assertStandardWifOK()

    async def test_balance(self):
        await self.assertBalanceOK()

    async def test_balances(self):
        await self.assertBalancesOK()

    async def test_unspent(self):
        await self.assertUnspentOK()

    async def test_unspents(self):
        await self.assertUnspentsOK()

    async def test_merkle_proof(self):
        await self.assertMerkleProofOK()

    async def test_history(self):
        await self.assertHistoryOK()

    async def test_histories(self):
        await self.assertHistoriesOK()

    async def test_balance_merkle_proven(self):
        await self.assertBalanceMerkleProvenOK()

    async def test_balances_merkle_proven(self):
        await self.assertBalancesMerkleProvenOK()

    async def test_block_header(self):
        await self.assertBlockHeaderOK()

    async def test_block_headers(self):
        await self.assertBlockHeadersOK()

    async def test_gettx(self):
        await self.assertGetTXOK()

    @unittest.skip("Intermittent failure")
    async def test_getverbosetx(self):
        await self.assertGetVerboseTXOK()

    async def test_gettxs(self):
        await self.assertGetTxsOK()

    async def test_transaction(self):
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertTransactionOK("812f055cc55515c2e1aa7a6aa38adc890d3c5d39781d38fb2e5f647b84b6b3fa")

    async def test_transaction_multisig(self):
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertMultiSigTransactionOK("a9b8184a05f317ba18f3e01482305bc6ba667f4a674e097de515d31381d902c1")

    async def test_sendmulti_recipient_tx(self):
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertSendMultiRecipientsTXOK("4646a5b2f143d98f474e3a60f745627a5ddbbccc7c1fceeeeeaa391c4c0b263d")

    async def test_send(self):
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertSendOK("01d0cb59eaaeaafc6d8e09d18bcd2b7beaf856936b41570819c4475d2dcdcd0e")

    async def test_subscribe_block_headers(self):
        await self.assertSubscribeBlockHeadersOK()

    async def test_subscribe_block_headers_sync(self):
        await self.assertSubscribeBlockHeadersSyncCallbackOK()

    async def test_latest_block(self):
        await self.assertLatestBlockOK()

    async def test_confirmations(self):
        await self.assertConfirmationsOK()




import unittest
from binascii import unhexlify
from cryptos import coins_async
from cryptos import cashaddr
from cryptos.testing.testcases_async import BaseAsyncCoinTestCase
from cryptos.electrumx_client.types import ElectrumXTx, ElectrumXMultiBalanceResponse
from cryptos.main import privtopub
from cryptos.types import TxInput
from typing import List, Type
from unittest import mock


class TestBitcoinCash(BaseAsyncCoinTestCase):
    name: str = "Bitcoin Cash"
    coin:  Type[coins_async.BaseCoin] = coins_async.BitcoinCash
    addresses = ["1MhTCMUjM1TEQ7RSwoCMVZy7ARa9aAP82Z",  ""
                 "18DEbpqpdmfNaosxwQhCNHDAChZYCNG836",
                 "175MvCJkNZT3zSdCntXj9vK7L6XKDWjLnD"]  # n2DQVQZiA2tVBDu4fNAjKVBS2RArWhfncv
    cash_addresses = ["bitcoincash:qr3sjptscfm7kqry6s67skm5dgsudwkmcsfvmsq7c6",
                      "bitcoincash:qp83jwvlc8clct6vpskr8jhyayr8u7ynhqf8z4glc3",
                      "bitcoincash:qpp28cg6sze9la3myp6v28ghg5fjhn9m5ynaj2uu6x"]
    multisig_addresses = ["35D72hVBzYXqNkyN63z28FHmSyPKuJh9Q2", "32tuh24PcKWQWfWitfp9NVhRuYjDKG7vCH"]
    cash_multisig_addresses = ["bitcoincash:pqnfj8jmtpj30fnjgc2gy0gs4l6sptdyhc84ukmr52",
                               "bitcoincash:pqxn06syr9twx9ecx892alre33yuuwn2gu7z0p7lzz"]
    privkeys: List[str] = ["098ddf01ebb71ead01fc52cb4ad1f5cafffb5f2d052dd233b3cad18e255e1db1",
                           "0861e1bb62504f5e9f03b59308005a6f2c12c34df108c6f7c52e5e712a08e91401",
                           "c396c62dfdc529645b822dc4eaa7b9ddc97dd8424de09ca19decce61e6732f71"]
    privkey_standard_wifs: List[str] = ['5HtVdWeLBUpSyqqSnnJoQnCiuc33Kg86i8V1gc1tKK13tw1Cqrg',
                                        "KwW1FKxkfefDyVStxvKH9qCCb9qaiFXBFZUy2mPLvTMap2f5YaXR",
                                        "5KJRe1vYXbE4hhvNjJjPX6iS1tqpksNKHChrQjzyYVDgh9Z8H5o"]
    fee: int = 500
    max_fee: int = fee
    testnet: bool = False
    unspent_addresses: List[str] = ["1KomPE4JdF7P4tBzb12cyqbBfrVp4WYxNS"]
    unspent_cash_addresses: List[str] = ["bitcoincash:qr8y5u55y3j9lsyk0rsmsvnm03udrnplcg6jpj24wk"]
    balance: ElectrumXMultiBalanceResponse = {'confirmed': 249077026, 'unconfirmed': 0}
    balances: List[ElectrumXMultiBalanceResponse] = [{'address': unspent_addresses[0]} | dict(balance)]
    unspent: List[ElectrumXTx] = [{'address': '1KomPE4JdF7P4tBzb12cyqbBfrVp4WYxNS',
                                   'height': 508381,
                                   'tx_hash': 'e3ead2c8e6ad22b38f49abd5ae7a29105f0f64d19865fd8ccb0f8d5b2665f476',
                                   'tx_pos': 1,
                                   'value': 249077026}]
    unspents: List[ElectrumXTx] = unspent
    history: List[ElectrumXTx] = [{'height': 508381, 'tx_hash': 'e3ead2c8e6ad22b38f49abd5ae7a29105f0f64d19865fd8ccb0f8d5b2665f476'}]
    histories: List[ElectrumXTx] = [{'address': unspent_addresses[0]} | dict(history[0])]
    min_latest_height: int = 129055
    txheight: int = 509045
    block_hash: str = "0000000000000000006d011e3ab462725dad9d4e8d1a7398bcc2895defd1fa3f"
    txinputs: List[TxInput] = [{'output': 'b3105972beef05e88cf112fd9718d32c270773462d62e4659dc9b4a2baafc038:0', 'value': 1157763509}]
    txid: str = "e3ead2c8e6ad22b38f49abd5ae7a29105f0f64d19865fd8ccb0f8d5b2665f476"
    raw_tx: str = "0200000001ab293b56edcc8d99b665ebe6265132408df132ecf7a1948d68bee425cef7bb63010000006b483045022100fca8f51dc515e85862cd087729136656e4f73f76eb1ce9d4ce90b092e4b9efea02204943929a08bab03e95dad6781e128f49a3d35af055a146a9c8e4aec3a4c90db54121039b190dc5e0bcea42cec072f7aebf097f379691b3dfcc67fd587dddc1d004eaa4feffffff0222ec4377000000001976a9149119c4f8dc64fde6e9d6f59ae9273993b858c03388ac229dd80e000000001976a914ce4a729424645fc09678e1b8327b7c78d1cc3fc288acdac10700"

    def test_cash_addr(self):
        # https://reference.cash/protocol/blockchain/encoding/cashaddr
        public_key_hash = unhexlify("211b74ca4686f81efda5641767fc84ef16dafe0b")
        addr = cashaddr.encode_full(self._coin.cash_hrp, 0, public_key_hash)
        self.assertEqual(addr, "bitcoincash:qqs3kax2g6r0s8ha54jpwelusnh3dkh7pvu23rzrru")

    def test_address_conversion(self):
        for addr, cashaddr in zip(self.addresses, self.cash_addresses):
            convert_cashaddr = self._coin.legacy_addr_to_cash_address(addr)
            self.assertEqual(convert_cashaddr, cashaddr)
            convert_addr = self._coin.cash_address_to_legacy_addr(cashaddr)
            self.assertEqual(addr, convert_addr)

    def test_cash_address_multisig_ok(self):
        pubs = [privtopub(priv) for priv in self.privkeys]
        script1, address1 = self._coin.mk_multsig_cash_address(*pubs, num_required=2)
        self.assertEqual(address1, self.cash_multisig_addresses[0])
        pubs2 = pubs[0:2]
        script2, address2 = self._coin.mk_multsig_cash_address(*pubs2)
        self.assertEqual(address2, self.cash_multisig_addresses[1])

    def test_address_conversion_multisig(self):
        for addr, cashaddr in zip(self.multisig_addresses, self.cash_multisig_addresses):
            convert_cashaddr = self._coin.legacy_addr_to_cash_address(addr)
            self.assertEqual(convert_cashaddr, cashaddr)
            convert_addr = self._coin.cash_address_to_legacy_addr(cashaddr)
            self.assertEqual(addr, convert_addr)

    def test_standard_wif_ok(self):
        self.assertStandardWifOK()
        for privkey, addr in zip(self.privkeys, self.cash_addresses):
            cash_addr = self._coin.privtocashaddress(privkey)
            self.assertEqual(cash_addr, addr)

    async def test_balance(self):
        await self.assertBalanceOK()

    @unittest.skip("Address needs updating")
    async def test_balance_cash_address(self):
        result = await self._coin.get_balance(self.unspent_cash_addresses[0])
        self.assertEqual(self.balance, result)

    async def test_balances(self):
        await self.assertBalancesOK()

    async def test_unspent(self):
        await self.assertUnspentOK()

    async def test_unspents(self):
        await self.assertUnspentsOK()

    async def test_merkle_proof(self):
        await self.assertMerkleProofOK()

    @unittest.skip("Intermittent Failures")
    async def test_history(self):
        await self.assertHistoryOK()

    async def test_histories(self):
        await self.assertHistoriesOK()

    @unittest.skip('Out of range error returned, not sure why')
    async def test_block_header(self):
        await self.assertBlockHeaderOK()

    async def test_block_headers(self):
        await self.assertBlockHeadersOK()

    async def test_gettx(self):
        await self.assertGetTXOK()

    @unittest.skip('Transaction not found')
    async def test_getverbosetx(self):
        await self.assertGetVerboseTXOK()

    async def test_gettxs(self):
        await self.assertTxsOK()

    @unittest.skip("Test address needs updating")
    async def test_balance_merkle_proven(self):
        await self.assertBalanceMerkleProvenOK()

    async def test_balances_merkle_proven(self):
        await self.assertBalancesMerkleProvenOK()

    async def test_transaction(self):
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertTransactionOK("744e0a28ed754ceb551a4fc3a45f57f0b3bd34ff415c3724a2ce774ccc24eff9")

    async def test_transaction_cash_address(self):
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertCashAddressTransactionOK(
                "744e0a28ed754ceb551a4fc3a45f57f0b3bd34ff415c3724a2ce774ccc24eff9")

    async def test_transaction_multisig(self):
        # 010000000176f465265b8d0fcb8cfd6598d1640f5f10297aaed5ab498fb322ade6c8d2eae301000000fd3d01004830450221009a556d396ad8ceec4cbb1cacda67933818deac5210a96a445ac13e97fe8c476102202682d3d25c62e23865b00e1c3ea36c53a2691c64dd43b04de76a35ecd64e5364414730440220045fd2daec62375ffe538c8758ab1461b4d7cc430ec24c57883a6048006daa1a022025d1dec4c9a36000539c283fb1be428b317b66e71ca29fa964efa4d287032f1b414ca9524104de476e251a827e58199ed4d6d7c2177f0a97a2dda150d7a9e59fc5682519eb94d37bc387edff66e7b0f16e92dd045fe968d63e1f203613b76ad733e5cdf8e818210391ed6bf1e0842997938ea2706480a7085b8bb253268fd12ea83a68509602b6e0410415991434e628402bebcbaa3261864309d2c6fd10c850462b9ef0258832822d35aa26e62e629d2337e3716784ca6c727c73e9600436ded7417d957318dc7a41eb53aeffffffff02914e6c070000000017a9140d37ea041956e3173831caaefc798c49ce3a6a4787514d6c070000000017a91426991e5b586517a6724614823d10aff500ada4be8700000000
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertMultiSigTransactionOK("9a44b27e7d03361ab66dff6bb9ce49a743b1e91123610dca94099ce92c50eeed")

    async def test_transaction_multisig_cash(self):
        # 010000000176f465265b8d0fcb8cfd6598d1640f5f10297aaed5ab498fb322ade6c8d2eae301000000fd3d0100473044022055dc9295816ce57fc78cd5e5b2e49d7811073e31532508a887e5acc562752e3a0220304788647824903c9fe0d10df3e0379391c85f572e15e966eac6744f3623473f41483045022100fa2a6222842a8102ab944cf9e0e282e907c3699921487f7da350a8879b45064c02202ab2a4eca7adb31c40b4210b8ff42171fad87370a8f84adbae5ceb694607f62e414ca9524104de476e251a827e58199ed4d6d7c2177f0a97a2dda150d7a9e59fc5682519eb94d37bc387edff66e7b0f16e92dd045fe968d63e1f203613b76ad733e5cdf8e818210391ed6bf1e0842997938ea2706480a7085b8bb253268fd12ea83a68509602b6e0410415991434e628402bebcbaa3261864309d2c6fd10c850462b9ef0258832822d35aa26e62e629d2337e3716784ca6c727c73e9600436ded7417d957318dc7a41eb53aeffffffff02914e6c070000000017a9140d37ea041956e3173831caaefc798c49ce3a6a4787117a6b070000000017a91426991e5b586517a6724614823d10aff500ada4be8700000000
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertCashAddressMultiSigTransactionOK("9a44b27e7d03361ab66dff6bb9ce49a743b1e91123610dca94099ce92c50eeed")


    async def test_sendmulti_recipient_tx(self):
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertSendMultiRecipientsTXOK("2f5046ed857685a08ffb642a8e352b7cdb5cc6b981d690086fb337b868f26dcd")

    async def test_send(self):
        with mock.patch('cryptos.electrumx_client.client.NotificationSession.send_request',
                        side_effect=self.mock_electrumx_send_request):
            await self.assertSendOK("9fea356516a2cb97b2242fb6b468f4f9264c26977e456b643cc787dc0f3480c0")

    async def test_subscribe_block_headers(self):
        await self.assertSubscribeBlockHeadersOK()

    async def test_subscribe_block_headers_sync(self):
        await self.assertSubscribeBlockHeadersSyncCallbackOK()

    async def test_latest_block(self):
        await self.assertLatestBlockOK()

    async def test_confirmations(self):
        await self.assertConfirmationsOK()



from .base import BaseCoin
from typing import Dict
from ..transaction import SIGHASH_ALL, SIGHASH_FORKID


class BitcoinCash(BaseCoin):
    coin_symbol = "BCH"
    display_name = "Bitcoin Cash"
    segwit_supported = False
    cash_address_supported = True
    magicbyte = 0
    script_magicbyte = 5
    wif_prefix = 0x80
    wif_script_types: Dict[str, int] = {
        'p2pkh': 0,
        'p2sh': 5,
    }
    hd_path = 145
    cash_hrp = "bitcoincash"
    hashcode = SIGHASH_ALL | SIGHASH_FORKID
    client_kwargs = {
        'server_file': 'bitcoin_cash.json',
    }
    minimum_fee = 500
    testnet_overrides = {
        'display_name': "Bitcoin Cash Testnet",
        'coin_symbol': "tbcc",
        'magicbyte': 111,
        'script_magicbyte': 196,
        'wif_prefix': 0xef,
        'cash_hrp': "bchtest",
        'xprv_headers': {
            'p2pkh': 0x04358394,
        },
        'xpub_headers': {
            'p2pkh': 0x043587cf,
        },
        'hd_path': 1,
        'client_kwargs': {
            'server_file': 'bitcoin_cash_testnet.json',
            'use_ssl': False
        },
    }

    def __init__(self, legacy=False, testnet=False, **kwargs):
        super(BitcoinCash, self).__init__(testnet=testnet, **kwargs)
        self.hd_path = 0 if legacy and testnet else self.hd_path



import unittest

from cryptos import coins_async
from cryptos.testing.testcases_async import BaseAsyncCoinTestCase
from cryptos.electrumx_client.types import ElectrumXTx, ElectrumXMultiBalanceResponse
from typing import List, Type


class TestBitcoinTestnet(BaseAsyncCoinTestCase):
    name: str = "Bitcoin Testnet"
    coin: Type[coins_async.BaseCoin] = coins_async.Bitcoin
    addresses: List[str] = ["n2DQVQZiA2tVBDu4fNAjKVBS2RArWhfncv",
                            "mnjBtsvoSo6dMvMaeyfaCCRV4hAF8WA2cu",
                            "mmbKDFPjBatJmZ6pWTW6yqXSC6826YLBX6"]
    segwit_addresses: List[str] = ["2N74sauceDn2qeHFJuNfJ3c1anxPcDRrVtz",
                                   "2NDpBxpK4obuGiFodKtYe3dXx14aPwDBPGU",
                                   "2Mt2f4knFtjLZz9CW2979Hw3tYiAYd6WcA1"]
    native_segwit_addresses: List[str] = ["tb1q95cgql39zvtc57g4vn8ytzmlvtt43skngdq0ue",
                                          "tb1qfuvnn87p787z7nqv9seu4e8fqel83yacg7yf2r",
                                          "tb1qst3pkm860tjt9y70ugnaluqyqnfa7h54ekyj66"]
    multisig_addresses: List[str] = ["2MvmK6SRDc13BaYbumBbtkCH2fKbViC5XEv",
                                     "2MtT7kkzRDn1kiT9GZoS1zSgh7twP145Qif"]
    native_segwit_multisig_addresses: List[str] = [
        "tb1q7e42a8gmgp5d7kw8myrruj5qnxp8edp7v5y0y95hrjz6t530uehqkj0tl2",
        "tb1qu7fz4uku8kh6tg7qghj7rnf88g858lal258gzfu85kx7vc5h0qpsyxrfnp"
    ]
    privkeys: List[str] = ["098ddf01ebb71ead01fc52cb4ad1f5cafffb5f2d052dd233b3cad18e255e1db1",
                           "0861e1bb62504f5e9f03b59308005a6f2c12c34df108c6f7c52e5e712a08e91401",
                           "c396c62dfdc529645b822dc4eaa7b9ddc97dd8424de09ca19decce61e6732f71"]
    public_keys: List[str] = [
        "04de476e251a827e58199ed4d6d7c2177f0a97a2dda150d7a9e59fc5682519eb94d37bc387edff66e7b0f16e92dd045fe968d63e1f203613b76ad733e5cdf8e818",
        "0391ed6bf1e0842997938ea2706480a7085b8bb253268fd12ea83a68509602b6e0",
        "0415991434e628402bebcbaa3261864309d2c6fd10c850462b9ef0258832822d35aa26e62e629d2337e3716784ca6c727c73e9600436ded7417d957318dc7a41eb"
    ]
    privkey_standard_wifs: List[str] = ['91f8DFTsmhtawuLjR8CiHNkgZGPkUqfJ45LxmENPf3k6fuX1m4N',
                                       'cMrziExc6iMV8vvAML8QX9hGDP8zNhcsKbdS9BqrRa1b4mhKvK6f',
                                       "9354Dkk67pJCfmRfMedJPhGPfZCXv2uWd9ZoVNMUtDxjUBbCVZK"]
    privkey_segwit_wifs: List[str] = ['cf4XyPZRSdJxnN3dKdaUodCn6HLiaiKUbTRNF8hbytcebJFytuLg',
                                       'cf2FrZgQT2Bm5xwhTUvC7VtEiFYJ3YAKeUshG6Y3QXX1dSAZ9s9h',
                                       "cmJAMcnywPbCT97PQHs1MWuuY3XD7uxE5S43K7wb43iR1Axqeupz"]
    privkey_native_segwit_wifs: List[str] = ["cWUuQECXGUPpor2rmZBb1T7Hqr94kJ3kS1oEggMWVQrwJy3wWMF4",
                                             "cWSdHQKWGsGd7SvvuQXJKKnkTpLeD7tbV3FZheBwv3mJM6yc95xc",
                                             "cciXnTS5mEg4Ud6crDU7ZLpRHcKZHVgVuzRukfbVZZxhiqfSyfBH"]
    fee: int = 1500
    max_fee: int = 3500
    testnet: bool = True
    min_latest_height: int = 1258030

    unspent_addresses: List[str] = ["ms31HApa3jvv3crqvZ3sJj7tC5TCs61GSA", "2MwHtiGJJqcFgNnbCu1REVy5ooDEeAAFXMy",
                                    "tb1qjap2aae2tsky3ctlh48yltev0sjdmx92yk76wq"]
    unspent: List[ElectrumXTx] = [{'tx_hash': '1d69dd7a23f18d86f514ff7d8ef85894ad00c61fb29f3f7597e9834ac2569c8c',
                                   'tx_pos': 0, 'height': 1238008, 'value': 180000000,
                                   'address': 'ms31HApa3jvv3crqvZ3sJj7tC5TCs61GSA'}]
    unspents: List[ElectrumXTx] = [
        {'tx_hash': '1d69dd7a23f18d86f514ff7d8ef85894ad00c61fb29f3f7597e9834ac2569c8c', 'tx_pos': 0, 'height': 1238008,
         'value': 180000000, 'address': 'ms31HApa3jvv3crqvZ3sJj7tC5TCs61GSA'},
        {'tx_hash': '70bd4ce0e4cf2977ab53e767865da21483977cdb94b1a36eb68d30829c9c392f', 'tx_pos': 1, 'height': 1275633,
         'value': 173980000, 'address': '2MwHtiGJJqcFgNnbCu1REVy5ooDEeAAFXMy'},
        {'tx_hash': '70bd4ce0e4cf2977ab53e767865da21483977cdb94b1a36eb68d30829c9c392f', 'tx_pos': 0, 'height': 1275633,
         'value': 6000000, 'address': 'tb1qjap2aae2tsky3ctlh48yltev0sjdmx92yk76wq'}]
    balance: ElectrumXMultiBalanceResponse = {'confirmed': 180000000, 'unconfirmed': 0}
    balances: List[ElectrumXMultiBalanceResponse] = [
        {'confirmed': 180000000, 'unconfirmed': 0, 'address': 'ms31HApa3jvv3crqvZ3sJj7tC5TCs61GSA'},
        {'confirmed': 173980000, 'unconfirmed': 0, 'address': '2MwHtiGJJqcFgNnbCu1REVy5ooDEeAAFXMy'},
        {'confirmed': 6000000, 'unconfirmed': 0, 'address': 'tb1qjap2aae2tsky3ctlh48yltev0sjdmx92yk76wq'}]
    history: List[ElectrumXTx] = [
        {'tx_hash': '1d69dd7a23f18d86f514ff7d8ef85894ad00c61fb29f3f7597e9834ac2569c8c', 'height': 1238008}]
    histories: List[ElectrumXTx] = [
        {'tx_hash': '1d69dd7a23f18d86f514ff7d8ef85894ad00c61fb29f3f7597e9834ac2569c8c', 'height': 1238008,
         'address': 'ms31HApa3jvv3crqvZ3sJj7tC5TCs61GSA'},
        {'tx_hash': 'e25d8f4036e44159b0364b45867e08ae47a57dda68ba800ba8abe1fb2dc54a40', 'height': 1275633,
         'address': '2MwHtiGJJqcFgNnbCu1REVy5ooDEeAAFXMy'},
        {'tx_hash': '70bd4ce0e4cf2977ab53e767865da21483977cdb94b1a36eb68d30829c9c392f', 'height': 1275633,
         'address': '2MwHtiGJJqcFgNnbCu1REVy5ooDEeAAFXMy'},
        {'tx_hash': '70bd4ce0e4cf2977ab53e767865da21483977cdb94b1a36eb68d30829c9c392f', 'height': 1275633,
         'address': 'tb1qjap2aae2tsky3ctlh48yltev0sjdmx92yk76wq'}]
    txid: str = "1d69dd7a23f18d86f514ff7d8ef85894ad00c61fb29f3f7597e9834ac2569c8c"
    txheight: int = 1238008
    block_hash: str = "00000000000ac694c157a56de45e2f985adefda11d3e2d7375905a03950852df"
    raw_tx: str = "01000000000101333ae299a6c6353d88e6540c67d8c82281259bc29b3313bcbc9b62a9a7e78a1b0100000017160014ffe21a1b058e7f8dedfcc3f1526f82303cff4fc7ffffffff020095ba0a000000001976a9147e585aa1913cf12e9948e90f67188ee9250d555688acfcb92b4d2c00000017a914e223701f10c2a5e7782ef6e10a2560f4c6e968a2870247304402207f2aa4118eee2ef231eab3afcbf6b01b4c1ca3672bd87a3864cf405741bd2c1d02202ab7502cbc50126f68cb2b366e5b3799d3ec0a3359c6a895a730a6891c7bcae10121023c13734016f27089393f9fc79736e4dca1f27725c68e720e1855202f3fbf037e00000000"

    def test_standard_wif_ok(self):
        self.assertStandardWifOK()

    def test_p2wpkh_p2sh_wif_ok(self):
        self.assertP2WPKH_P2SH_WifOK()

    def test_p2wpkh_wif_ok(self):
        self.assertP2WPKH_WIFOK()

    async def test_balance(self):
        await self.assertBalanceOK()

    async def test_balances(self):
        await self.assertBalancesOK()

    async def test_unspent(self):
        await self.assertUnspentOK()

    async def test_unspents(self):
        await self.assertUnspentsOK()

    async def test_merkle_proof(self):
        await self.assertMerkleProofOK()

    async def test_history(self):
        await self.assertHistoryOK()

    async def test_histories(self):
        await self.assertHistoriesOK()

    async def test_balance_merkle_proven(self):
        await self.assertBalanceMerkleProvenOK()

    async def test_balances_merkle_proven(self):
        await self.assertBalancesMerkleProvenOK()

    async def test_block_header(self):
        await self.assertBlockHeaderOK()

    async def test_block_headers(self):
        await self.assertBlockHeadersOK()

    async def test_gettx(self):
        await self.assertGetSegwitTXOK()

    @unittest.skip('Intermittent failures')
    async def test_getverbosetx(self):
        await self.assertGetVerboseTXOK()

    async def test_gettxs(self):
        await self.assertGetSegwitTxsOK()

    async def test_transaction(self):
        """
        Sample transaction:
        TxID: ef60508ebe9f684fa881ad41ba278365f238a5fca36af78f3f106c30f7020eca
        """
        await self.assertTransactionOK()

    async def test_transaction_p2pk(self):
        """
        Sample transaction:
        TxID:
        """
        await self.assertTransactionToPKOK()

    async def test_transaction_segwit(self):
        """
        Sample transaction:
        TxID: 0494107dea76e27658b245744ffda3766010af347c3dd68abb2686ca080d1cd5
        """
        await self.assertSegwitTransactionOK()

    async def test_transaction_native_segwit(self):
        """
        Sample transaction:
        TxID: b368c55cd8895c561f176645ce418b3e3ac7e3c98aa74f4a7bcf286e53c86515
        """
        await self.assertNativeSegwitTransactionOK()

    async def test_transaction_mixed_segwit(self):
        """
        Sample transaction:
        TxID: 15470f798231153ce9d6427cd223c8bd591b2ebf6c059e19347db4df038d6e86
        """
        await self.assertMixedSegwitTransactionOK()

    @unittest.skip('Intermittent failure')
    async def test_transaction_multisig(self):
        """
        Sample transaction:
        TxID: a555123e8e49e32e8d462705f278c5cd5a3d08ea5e7e738ecc286dae1a1eac38
        """
        await self.assertMultiSigTransactionOK()

    async def test_transaction_native_segwit_multisig(self):
        """
        Sample transaction:
        TxID: b710704a7939e3e0c82e643faaa3a602549c416a2004feba13e2ef7a9e95dddb
        """
        await self.assertNativeSegwitMultiSigTransactionOK()

    async def test_sendmulti_recipient_tx(self):
        """
        Sample transaction:
        TxID: 8b364c9337998c2586350804553f0f66f97372fe99ed0506a279e4e344495fb8
        """
        await self.assertSendMultiRecipientsTXOK()

    async def test_send(self):
        """
        Sample transaction:
        TxID: 21fdea144f28d0cb1da99c5fe7c96268aaa98ddef3a14fd627b44ea31ce0be3e
        """
        await self.assertSendOK()

    async def test_subscribe_block_headers(self):
        await self.assertSubscribeBlockHeadersOK()

    async def test_subscribe_block_headers_sync(self):
        await self.assertSubscribeBlockHeadersSyncCallbackOK()

    async def test_latest_block(self):
        await self.assertLatestBlockOK()

    async def test_confirmations(self):
        await self.assertConfirmationsOK()

    @unittest.skip('Intermittent failures')
    async def test_subscribe_address(self):
        await self.assertSubscribeAddressOK()

    async def test_subscribe_address_sync(self):
        await self.assertSubscribeAddressSyncCallbackOK()

    async def test_subscribe_address_transactions(self):
        await self.assertSubscribeAddressTransactionsOK()

    async def test_subscribe_address_transactions_sync(self):
        await self.assertSubscribeAddressTransactionsSyncOK()



import unittest

from cryptos import coins_async
from cryptos.main import privtopub
from cryptos.testing.testcases_async import BaseAsyncCoinTestCase
from cryptos.types import ElectrumXTx, TxOut
from cryptos.electrumx_client.types import ElectrumXMultiBalanceResponse
from typing import List, Type


class TestBitcoinCashTestnet(BaseAsyncCoinTestCase):
    name = "Bitcoin Cash Testnet"
    coin: Type[coins_async.BaseCoin] = coins_async.BitcoinCash
    addresses: List[str] = ["n2DQVQZiA2tVBDu4fNAjKVBS2RArWhfncv",
                            "mnjBtsvoSo6dMvMaeyfaCCRV4hAF8WA2cu",
                            "mmbKDFPjBatJmZ6pWTW6yqXSC6826YLBX6"]
    cash_addresses = ["bchtest:qr3sjptscfm7kqry6s67skm5dgsudwkmcsd7lhzflx",
                      "bchtest:qp83jwvlc8clct6vpskr8jhyayr8u7ynhqd4xj2gld",
                      "bchtest:qpp28cg6sze9la3myp6v28ghg5fjhn9m5yh0kd7ta6"]
    privkeys: List[str] = [
        "098ddf01ebb71ead01fc52cb4ad1f5cafffb5f2d052dd233b3cad18e255e1db1",
        "0861e1bb62504f5e9f03b59308005a6f2c12c34df108c6f7c52e5e712a08e91401",
        "c396c62dfdc529645b822dc4eaa7b9ddc97dd8424de09ca19decce61e6732f71"]  # Private keys for above address_derivations in same order
    privkey_standard_wifs: List[str] = ['91f8DFTsmhtawuLjR8CiHNkgZGPkUqfJ45LxmENPf3k6fuX1m4N',
                                       'cMrziExc6iMV8vvAML8QX9hGDP8zNhcsKbdS9BqrRa1b4mhKvK6f',
                                       "9354Dkk67pJCfmRfMedJPhGPfZCXv2uWd9ZoVNMUtDxjUBbCVZK"]
    multisig_addresses: List[str] = ["2MvmK6SRDc13BaYbumBbtkCH2fKbViC5XEv", "2MtT7kkzRDn1kiT9GZoS1zSgh7twP145Qif"]
    cash_multisig_addresses: List[str] = ["bchtest:pqnfj8jmtpj30fnjgc2gy0gs4l6sptdyhcr8c3e5nk",
                                          "bchtest:pqxn06syr9twx9ecx892alre33yuuwn2gu6stxug97"]
    fee: int = 1000
    max_fee: int = 3500
    testnet: bool = True

    min_latest_height: int = 1524427
    unspent_addresses: List[str] = ["ms31HApa3jvv3crqvZ3sJj7tC5TCs61GSA"]
    unspent_cash_addresses: List[str] = ['bchtest:qpl9sk4pjy70zt5efr5s7ecc3m5j2r242c4czjmhfy']
    unspent: List[ElectrumXTx] = [{'address': 'ms31HApa3jvv3crqvZ3sJj7tC5TCs61GSA',
                                   'height': 1196454,
                                   'tx_hash': '80700e6d1125deafa22b307f6c7c99e75771f9fc05517fc795a1344eca7c8472',
                                   'tx_pos': 0,
                                   'value': 550000000}]
    unspents: List[ElectrumXTx] = unspent
    balance: ElectrumXMultiBalanceResponse = {'confirmed': 550000000, 'unconfirmed': 0}
    balances: List[ElectrumXMultiBalanceResponse] = [{'address': unspent_addresses[0]} | dict(balance)]
    history: List[ElectrumXTx] = [{'height': 1196454, 'tx_hash': '80700e6d1125deafa22b307f6c7c99e75771f9fc05517fc795a1344eca7c8472'}]
    histories: List[ElectrumXTx] = [{'address': unspent_addresses[0]} | dict(history[0])]
    txid: str = "b4dd5908cca851d861b9d2ca267a901bb6f581f2bb096fbf42a28cc2d98e866a"
    txheight: int = 1196454
    block_hash: str = "000000002bab447cbd0c60829a80051e320aa6308d578db3369eb85b2ebb9f46"
    txinputs: List[TxOut] = [{'output': "cbd43131ee11bc9e05f36f55088ede26ab5fb160cc3ff11785ce9cc653aa414b:1", 'value': 96190578808}]
    raw_tx: str = "01000000014b41aa53c69cce8517f13fcc60b15fab26de8e08556ff3059ebc11ee3131d4cb010000006b483045022100b9050a1d58f36a771c4e0869900fb0474b809b134fdad566742e5b3a0ed7580d022065b80e9cc2bc9b921a9b0aad12228d9967345959b021214dbe60b3ffa44dbf0e412102ae83c12f8e2a686fb6ebb25a9ebe39fcd71d981cc6c172fedcdd042536a328f2ffffffff0200ab9041000000001976a914c384950342cb6f8df55175b48586838b03130fad88acd88ed523160000001976a9143479daa7de5c6d8dad24535e648861d4e7e3f7e688ac00000000"

    def test_address_conversion(self):
        for addr, cashaddr in zip(self.addresses, self.cash_addresses):
            convert_cashaddr = self._coin.legacy_addr_to_cash_address(addr)
            self.assertEqual(convert_cashaddr, cashaddr)
            convert_addr = self._coin.cash_address_to_legacy_addr(cashaddr)
            self.assertEqual(addr, convert_addr)

    def test_cash_address_multisig_ok(self):
        pubs = [privtopub(priv) for priv in self.privkeys]
        script1, address1 = self._coin.mk_multsig_cash_address(*pubs, num_required=2)
        self.assertEqual(address1, self.cash_multisig_addresses[0])
        pubs2 = pubs[0:2]
        script2, address2 = self._coin.mk_multsig_cash_address(*pubs2)
        self.assertEqual(address2, self.cash_multisig_addresses[1])

    def test_address_conversion_multisig(self):
        for addr, cashaddr in zip(self.multisig_addresses, self.cash_multisig_addresses):
            convert_cashaddr = self._coin.legacy_addr_to_cash_address(addr)
            self.assertEqual(convert_cashaddr, cashaddr)
            convert_addr = self._coin.cash_address_to_legacy_addr(cashaddr)
            self.assertEqual(addr, convert_addr)

    def test_standard_wif_ok(self):
        self.assertStandardWifOK()
        for privkey, addr in zip(self.privkeys, self.cash_addresses):
            cash_addr = self._coin.privtocashaddress(privkey)
            self.assertEqual(cash_addr, addr)

    async def test_balance(self):
        await self.assertBalanceOK()

    async def test_balance_cash_address(self):
        result = await self._coin.get_balance(self.unspent_cash_addresses[0])
        self.assertEqual(self.balance, result)

    async def test_balances(self):
        await self.assertBalancesOK()

    async def test_unspent(self):
        await self.assertUnspentOK()

    async def test_unspents(self):
        await self.assertUnspentsOK()

    async def test_merkle_proof(self):
        await self.assertMerkleProofOK()

    async def test_history(self):
        await self.assertHistoryOK()

    async def test_histories(self):
        await self.assertHistoriesOK()

    async def test_block_header(self):
        await self.assertBlockHeaderOK()

    async def test_block_headers(self):
        await self.assertBlockHeadersOK()

    async def test_gettx(self):
        await self.assertGetTXOK()

    @unittest.skip('Intermittent failures')
    async def test_getverbosetx(self):
        await self.assertGetVerboseTXOK()

    async def test_gettxs(self):
        await self.assertGetTxsOK()

    async def test_balance_merkle_proven(self):
        await self.assertBalanceMerkleProvenOK()

    async def test_balances_merkle_proven(self):
        await self.assertBalancesMerkleProvenOK()

    async def test_transaction(self):
        """
        Sample transaction:
        TxID: d4e8e93ba458c675270a5e6ac6772e35356ec95c37f8de6eb4a7a74103ecac8a
        """
        await self.assertTransactionOK()

    async def test_transaction_cash_address(self):
        """
        Sample transaction:
        TxID: 1ec96ce25a0104cda556f16d0d630768308a6b14dd35363dabe62fa96aa3237a
        """
        await self.assertCashAddressTransactionOK()

    async def test_transaction_multisig(self):
        """
        Sample transaction:
        TxID: c8987d59357f108fff46837b9309b28a1dc91d0fa4daa2c2f515107f61943a05
        """
        await self.assertMultiSigTransactionOK()

    async def test_transaction_multisig_cash_address(self):
        """
        Sample transaction:
        TxID: eb8a0ad9434786bfe69c992f286cbe391c4845058ae130f8a35e205e3280c6d4
        """
        await self.assertCashAddressMultiSigTransactionOK()

    async def test_sendmulti_recipient_tx(self):
        """
        Sample transaction:
        TxID: 49a731c4eaae1c6570590cd3eb5f4af4d4f6b282186b368b574169e9a7d576ab
        """
        await self.assertSendMultiRecipientsTXOK()

    async def test_send(self):
        """
        Sample transaction:
        TxID: ae9dd16e61521791659080c58299937a5d0ac2d20608e8bc1f494a08f6d9f5fb
        """
        await self.assertSendOK()

    async def test_subscribe_block_headers(self):
        await self.assertSubscribeBlockHeadersOK()

    async def test_subscribe_block_headers_sync(self):
        await self.assertSubscribeBlockHeadersSyncCallbackOK()

    async def test_latest_block(self):
        await self.assertLatestBlockOK()

    async def test_confirmations(self):
        await self.assertConfirmationsOK()

    async def test_subscribe_address(self):
        await self.assertSubscribeAddressOK()

    async def test_subscribe_address_sync(self):
        await self.assertSubscribeAddressSyncCallbackOK()

    async def test_subscribe_address_transactions(self):
        await self.assertSubscribeAddressTransactionsOK()

    async def test_subscribe_address_transactions_sync(self):
        await self.assertSubscribeAddressTransactionsSyncOK()



import unittest
from unittest import mock

from cryptos import *


# TODO passphrase/seed_extension
class TestWalletKeystoreAddressIntegrity(unittest.TestCase):

    gap_limit = 1  # make tests run faster

    def _check_seeded_keystore_sanity(self, ks):
        self.assertTrue (ks.is_deterministic())
        self.assertFalse(ks.is_watching_only())
        self.assertFalse(ks.can_import())
        self.assertTrue (ks.has_seed())

    def _check_xpub_keystore_sanity(self, ks):
        self.assertTrue (ks.is_deterministic())
        self.assertTrue (ks.is_watching_only())
        self.assertFalse(ks.can_import())
        self.assertFalse(ks.has_seed())

    def _create_standard_wallet(self, seed, coin):
        w = coin.wallet(seed)
        return w

    def _create_multisig_wallet(self, ks1, ks2, ks3=None):
        """Creates a 2-of-2 or 2-of-3 multisig wallet."""
        store = storage.WalletStorage('if_this_exists_mocking_failed_648151893')
        store.put('x%d/' % 1, ks1.dump())
        store.put('x%d/' % 2, ks2.dump())
        if ks3 is None:
            multisig_type = "%dof%d" % (2, 2)
        else:
            multisig_type = "%dof%d" % (2, 3)
            store.put('x%d/' % 3, ks3.dump())
        store.put('wallet_type', multisig_type)
        store.put('gap_limit', self.gap_limit)
        w = wallet.Multisig_Wallet(store)
        w.synchronize()
        return w

    def test_electrum_seed_standard(self):
        seed_words = 'cycle rocket west magnet parrot shuffle foot correct salt library feed song'
        self.assertEqual(seed_type(seed_words), 'standard')

        w = Bitcoin().electrum_wallet(seed_words)

        self._check_seeded_keystore_sanity(w.keystore)
        self.assertTrue(isinstance(w.keystore, keystore.BIP32_KeyStore))

        self.assertEqual(w.keystore.xtype, 'p2pkh')

        self.assertEqual(w.keystore.xprv, 'xprv9s21ZrQH143K32jECVM729vWgGq4mUDJCk1ozqAStTphzQtCTuoFmFafNoG1g55iCnBTXUzz3zWnDb5CVLGiFvmaZjuazHDL8a81cPQ8KL6')
        self.assertEqual(w.keystore.xpub, 'xpub661MyMwAqRbcFWohJWt7PHsFEJfZAvw9ZxwQoDa4SoMgsDDM1T7WK3u9E4edkC4ugRnZ8E4xDZRpk8Rnts3Nbt97dPwT52CwBdDWroaZf8U')

        self.assertEqual(w.new_receiving_address(), '1NNkttn1YvVGdqBW4PR6zvc3Zx3H5owKRf')
        self.assertEqual(w.new_change_address(), '1KSezYMhAJMWqFbVFB2JshYg69UpmEXR4D')

        self.assertEqual(w.receiving_addresses, ['1NNkttn1YvVGdqBW4PR6zvc3Zx3H5owKRf'])
        self.assertEqual(w.change_addresses, ['1KSezYMhAJMWqFbVFB2JshYg69UpmEXR4D'])

        self.assertEqual(w.new_receiving_addresses(2), ['18KCZB4y7SKPmTGN2rDQuUsNxyVBX7CAbG', '15NuNdCwazxBd3y2MZin9qevuviA3pJe65'])
        self.assertEqual(w.new_change_addresses(2), ['1GSmpT2UWuELEMEv6GmoAYZwmUY1oFsypn', '1B6yGxiPdW8y8Zp8TLVSFV2jdKDiy3fDWv'])

        self.assertEqual(w.privkey('1NNkttn1YvVGdqBW4PR6zvc3Zx3H5owKRf', formt="wif_compressed"), "L4xstkeBfS6RbE6FezmKsHheFqktMMHQMsK7D1Hr5dvFbNhnekHM")
        self.assertEqual(w.privkey('1KSezYMhAJMWqFbVFB2JshYg69UpmEXR4D', formt="wif_compressed"), "L3eak2Wbw2MVgJBTcjZwNz7Ty6nw9YbDEbzAXpPxfUEfpLfXbVkA")
        self.assertEqual(w.privkey('18KCZB4y7SKPmTGN2rDQuUsNxyVBX7CAbG', formt="wif_compressed"), "KxERQEvau1865MuqPa1HdhKGMHAgsQJUq5XczC4rijEp7xj56hui")
        self.assertEqual(w.privkey('1GSmpT2UWuELEMEv6GmoAYZwmUY1oFsypn', formt="wif_compressed"), "L35MWQyAYLjURUPcSYvFnVj4KwxXXL1PKkqdnPwrfT95a5QtFMhv")
        self.assertEqual(w.privkey('15NuNdCwazxBd3y2MZin9qevuviA3pJe65', formt="wif_compressed"), "L46AzQn5yezvjNzgsNAPxZHCxm2rFXxDjrMzcnmBm1RJs4Gfp52P")
        self.assertEqual(w.privkey('1B6yGxiPdW8y8Zp8TLVSFV2jdKDiy3fDWv', formt="wif_compressed"), "L4miHgBAdGnpTdXi2psXN1TUoXNASfzrF7RGQqVfzbUgzV5vmfzJ")

    def test_electrum_seed_standard_watch(self):
        xpub = 'xpub661MyMwAqRbcFWohJWt7PHsFEJfZAvw9ZxwQoDa4SoMgsDDM1T7WK3u9E4edkC4ugRnZ8E4xDZRpk8Rnts3Nbt97dPwT52CwBdDWroaZf8U'

        w = Bitcoin().watch_electrum_wallet(xpub)

        self.assertTrue(isinstance(w.keystore, keystore.BIP32_KeyStore))
        self.assertEqual(w.keystore.xtype, 'p2pkh')
        self.assertTrue(w.is_watching_only)

        self.assertEqual(w.new_receiving_address(), '1NNkttn1YvVGdqBW4PR6zvc3Zx3H5owKRf')
        self.assertEqual(w.new_change_address(), '1KSezYMhAJMWqFbVFB2JshYg69UpmEXR4D')

        self.assertEqual(w.receiving_addresses, ['1NNkttn1YvVGdqBW4PR6zvc3Zx3H5owKRf'])
        self.assertEqual(w.change_addresses, ['1KSezYMhAJMWqFbVFB2JshYg69UpmEXR4D'])

        self.assertEqual(w.new_receiving_addresses(2), ['18KCZB4y7SKPmTGN2rDQuUsNxyVBX7CAbG', '15NuNdCwazxBd3y2MZin9qevuviA3pJe65'])
        self.assertEqual(w.new_change_addresses(2), ['1GSmpT2UWuELEMEv6GmoAYZwmUY1oFsypn', '1B6yGxiPdW8y8Zp8TLVSFV2jdKDiy3fDWv'])

    def test_electrum_seed_standard_testnet(self):
        seed_words = 'cycle rocket west magnet parrot shuffle foot correct salt library feed song'
        self.assertEqual(seed_type(seed_words), 'standard')

        w = Bitcoin(testnet=True).electrum_wallet(seed_words)

        self._check_seeded_keystore_sanity(w.keystore)
        self.assertTrue(isinstance(w.keystore, keystore.BIP32_KeyStore))

        self.assertEqual(w.keystore.xtype, 'p2pkh')

        self.assertEqual(w.keystore.xprv, 'xprv9s21ZrQH143K32jECVM729vWgGq4mUDJCk1ozqAStTphzQtCTuoFmFafNoG1g55iCnBTXUzz3zWnDb5CVLGiFvmaZjuazHDL8a81cPQ8KL6')
        self.assertEqual(w.keystore.xpub, 'xpub661MyMwAqRbcFWohJWt7PHsFEJfZAvw9ZxwQoDa4SoMgsDDM1T7WK3u9E4edkC4ugRnZ8E4xDZRpk8Rnts3Nbt97dPwT52CwBdDWroaZf8U')

        self.assertEqual(w.new_receiving_address(), 'n2tiBwrzMwvXQwf7mxPUpqpNRwdyxwbNqT')
        self.assertEqual(w.new_change_address(), 'myxcHbSfyKnmcN56xjzghckzx95XfYa75v')

        self.assertEqual(w.receiving_addresses, ['n2tiBwrzMwvXQwf7mxPUpqpNRwdyxwbNqT'])
        self.assertEqual(w.change_addresses, ['myxcHbSfyKnmcN56xjzghckzx95XfYa75v'])

        self.assertEqual(w.new_receiving_addresses(2), ['mnq9rE9wvTkeYZjykRBnjQ5hpy5tQFS2VC', 'mjtrfgHvQ2PSQASe58h9yksFmvJrsmx3i7'])
        self.assertEqual(w.new_change_addresses(2), ['mvxj7W7TKvfb1TiXoqkAzTnGdU8ie5Rmrv', 'mqcva1oNSXaDugHkAuTp5QF4VJpRroBguF'])

        self.assertEqual(w.receiving_address(100), "moopS26mFMQdeyu8ycPQFXZS5mSYAbBT6r")

        self.assertEqual(w.privkey('n2tiBwrzMwvXQwf7mxPUpqpNRwdyxwbNqT', formt="wif_compressed"), "cVKsMfe36VngkfZX3QaTEcCht54J1oP6RuTaKRkMakaFr7hRcH9c")
        self.assertEqual(w.privkey('myxcHbSfyKnmcN56xjzghckzx95XfYa75v', formt="wif_compressed"), "cU1aCwWTN63kqjej19P4kJcXbL6LozguJe8deErUAatg55jcrjW4")
        self.assertEqual(w.privkey('mnq9rE9wvTkeYZjykRBnjQ5hpy5tQFS2VC', formt="wif_compressed"), "cNbQs9vSL4pMEoP6mypR11pKyWU6XrQAu7g66cXNDqtpNhqNqKzr")
        self.assertEqual(w.privkey('mjtrfgHvQ2PSQASe58h9yksFmvJrsmx3i7', formt="wif_compressed"), "cUTATKmwQihBtpTxFmyXKsnGazLFuz3uotWTjDDhG85K7oR9mKQs")
        self.assertEqual(w.privkey('mvxj7W7TKvfb1TiXoqkAzTnGdU8ie5Rmrv', formt="wif_compressed"), "cTSLyKy1yQRjaurspxjP9pE7xBFwBn75Pnz6tpQNAZo5ppZ2qPvs")
        self.assertEqual(w.privkey('mqcva1oNSXaDugHkAuTp5QF4VJpRroBguF', formt="wif_compressed"), "cV8hkbB24LV5d4zyREgejKxYRkfa786YK9ZjXFxBVi8hFE88Uqts")
        self.assertEqual(w.privkey('moopS26mFMQdeyu8ycPQFXZS5mSYAbBT6r', formt="wif_compressed"), "cRCMQHdaJeA15bvRuHkPqzk46uikwpuzPUtLkMcraUHnx98BVnFH")

    def test_electrum_seed_segwit(self):
        seed_words = 'bitter grass shiver impose acquire brush forget axis eager alone wine silver'
        self.assertEqual(seed_type(seed_words), 'segwit')

        w = Bitcoin().electrum_wallet(seed_words)

        self._check_seeded_keystore_sanity(w.keystore)
        self.assertTrue(isinstance(w.keystore, keystore.BIP32_KeyStore))

        self.assertEqual(w.keystore.xtype, 'p2wpkh')

        self.assertEqual(w.keystore.xprv, 'zprvAZswDvNeJeha8qZ8g7efN3FXYVJLaEUsE9TW6qXDEbVe74AZ75c2sZFZXPNFzxnhChDQ89oC8C5AjWwHmH1HeRKE1c4kKBQAmjUDdKDUZw2')
        self.assertEqual(w.keystore.xpub, 'zpub6nsHdRuY92FsMKdbn9BfjBCG6X8pyhCibNP6uDvpnw2cyrVhecvHRMa3Ne8kdJZxjxgwnpbHLkcR4bfnhHy6auHPJyDTQ3kianeuVLdkCYQ')

        self.assertEqual(w.new_receiving_address(), 'bc1q3g5tmkmlvxryhh843v4dz026avatc0zzr6h3af')
        self.assertEqual(w.new_change_address(), 'bc1qdy94n2q5qcp0kg7v9yzwe6wvfkhnvyzje7nx2p')

        self.assertEqual(w.receiving_addresses, ['bc1q3g5tmkmlvxryhh843v4dz026avatc0zzr6h3af'])
        self.assertEqual(w.change_addresses, ['bc1qdy94n2q5qcp0kg7v9yzwe6wvfkhnvyzje7nx2p'])

        self.assertEqual(w.new_receiving_addresses(2), ['bc1q9pzjpjq4nqx5ycnywekcmycqz0wjp2nq604y2n', 'bc1qk8pyjanxrtv9039mz3muzjalfuk9r5xugm3803'])
        self.assertEqual(w.new_change_addresses(2), ['bc1q6xwxcw6m9ga35687tnu5tstmsvmzjwdnzktemv', 'bc1qgl92qp6haea2ktxe93umzsyhv9z8u32xtvk3qk'])

        self.assertEqual(w.receiving_address(100), "bc1ql6m8hpkst40p5g3fp6vkc69fysn8w9cpanuh09")

        self.assertEqual(w.privkey('bc1q3g5tmkmlvxryhh843v4dz026avatc0zzr6h3af', formt="wif_compressed"), "L9fSXYNxYWHJWUqrQ6yhZCAJXq6XsfvcJ1Y2EnMAZfLLRNVQswQj")
        self.assertEqual(w.privkey('bc1qdy94n2q5qcp0kg7v9yzwe6wvfkhnvyzje7nx2p', formt="wif_compressed"), "L8rPGyfyzdLLEzxuBeC87Jvpp8FKxwrRtmkZ2PkRmRjqxNF8TVwG")
        self.assertEqual(w.privkey('bc1q9pzjpjq4nqx5ycnywekcmycqz0wjp2nq604y2n', formt="wif_compressed"), "LDSB7EmiqX25GkFpzJi7cnNBhpyrUaGLSJiqGu8oKJcH7q91dwaL")
        self.assertEqual(w.privkey('bc1qk8pyjanxrtv9039mz3muzjalfuk9r5xugm3803', formt="wif_compressed"), "LD8dYgymFBKz2mju2KsaKHe2AuisXCTSgaiRkHaVbrzuUsNkA24U")
        self.assertEqual(w.privkey('bc1q6xwxcw6m9ga35687tnu5tstmsvmzjwdnzktemv', formt="wif_compressed"), "L7NeR6r9yU2n4zddxTCUpKYmzugYuouyLsCZR9naTqkBW6sjpxDM")
        self.assertEqual(w.privkey('bc1qgl92qp6haea2ktxe93umzsyhv9z8u32xtvk3qk', formt="wif_compressed"), "L6mxN5yrGMdbVFM7ihxFMUi7fsj9gmC3GFqsCa2an25SKdyqpUyR")
        self.assertEqual(w.privkey('bc1ql6m8hpkst40p5g3fp6vkc69fysn8w9cpanuh09', formt="wif_compressed"), "LAc21LGaLeN8bfAFwGHUFqPbu1cazb53ycaHWRBykrpwoVf5ZC3N")

    def test_electrum_seed_segwit_watch(self):
        xpub = 'zpub6nsHdRuY92FsMKdbn9BfjBCG6X8pyhCibNP6uDvpnw2cyrVhecvHRMa3Ne8kdJZxjxgwnpbHLkcR4bfnhHy6auHPJyDTQ3kianeuVLdkCYQ'

        w = Bitcoin().watch_electrum_p2wpkh_wallet(xpub)

        self.assertTrue(isinstance(w.keystore, keystore.BIP32_KeyStore))
        self.assertEqual(w.keystore.xtype, 'p2wpkh')
        self.assertTrue(w.is_watching_only)

        self.assertEqual(w.new_receiving_address(), 'bc1q3g5tmkmlvxryhh843v4dz026avatc0zzr6h3af')
        self.assertEqual(w.new_change_address(), 'bc1qdy94n2q5qcp0kg7v9yzwe6wvfkhnvyzje7nx2p')

        self.assertEqual(w.receiving_addresses, ['bc1q3g5tmkmlvxryhh843v4dz026avatc0zzr6h3af'])
        self.assertEqual(w.change_addresses, ['bc1qdy94n2q5qcp0kg7v9yzwe6wvfkhnvyzje7nx2p'])

        self.assertEqual(w.new_receiving_addresses(2), ['bc1q9pzjpjq4nqx5ycnywekcmycqz0wjp2nq604y2n', 'bc1qk8pyjanxrtv9039mz3muzjalfuk9r5xugm3803'])
        self.assertEqual(w.new_change_addresses(2), ['bc1q6xwxcw6m9ga35687tnu5tstmsvmzjwdnzktemv', 'bc1qgl92qp6haea2ktxe93umzsyhv9z8u32xtvk3qk'])

        self.assertEqual(w.receiving_address(100), "bc1ql6m8hpkst40p5g3fp6vkc69fysn8w9cpanuh09")

    def test_electrum_seed_segwit_testnet(self):
        seed_words = 'bitter grass shiver impose acquire brush forget axis eager alone wine silver'
        self.assertEqual(seed_type(seed_words), 'segwit')

        w = Bitcoin(testnet=True).electrum_wallet(seed_words)

        self._check_seeded_keystore_sanity(w.keystore)
        self.assertTrue(isinstance(w.keystore, keystore.BIP32_KeyStore))

        self.assertEqual(w.keystore.xtype, 'p2wpkh')

        self.assertEqual(w.keystore.xprv, 'zprvAZswDvNeJeha8qZ8g7efN3FXYVJLaEUsE9TW6qXDEbVe74AZ75c2sZFZXPNFzxnhChDQ89oC8C5AjWwHmH1HeRKE1c4kKBQAmjUDdKDUZw2')
        self.assertEqual(w.keystore.xpub, 'zpub6nsHdRuY92FsMKdbn9BfjBCG6X8pyhCibNP6uDvpnw2cyrVhecvHRMa3Ne8kdJZxjxgwnpbHLkcR4bfnhHy6auHPJyDTQ3kianeuVLdkCYQ')

        self.assertEqual(w.new_receiving_address(), 'tb1q3g5tmkmlvxryhh843v4dz026avatc0zzfuvzx6')
        self.assertEqual(w.new_change_address(), 'tb1qdy94n2q5qcp0kg7v9yzwe6wvfkhnvyzjncg43j')

        self.assertEqual(w.receiving_addresses, ['tb1q3g5tmkmlvxryhh843v4dz026avatc0zzfuvzx6'])
        self.assertEqual(w.change_addresses, ['tb1qdy94n2q5qcp0kg7v9yzwe6wvfkhnvyzjncg43j'])

        self.assertEqual(w.new_receiving_addresses(2), ['tb1q9pzjpjq4nqx5ycnywekcmycqz0wjp2nqsfwh3q', 'tb1qk8pyjanxrtv9039mz3muzjalfuk9r5xuza255z'])
        self.assertEqual(w.new_change_addresses(2), ['tb1q6xwxcw6m9ga35687tnu5tstmsvmzjwdngss2ql', 'tb1qgl92qp6haea2ktxe93umzsyhv9z8u32xp2dzm9'])

        self.assertEqual(w.receiving_address(100), "tb1ql6m8hpkst40p5g3fp6vkc69fysn8w9cph48y5k")

        self.assertEqual(w.privkey('tb1q3g5tmkmlvxryhh843v4dz026avatc0zzfuvzx6', formt="wif_compressed"), "ca2RzTNoyZyZfvK7nWnpvWfNA4PwY82JN3gVMCog4mzLg7ZXqxBk")
        self.assertEqual(w.privkey('tb1q9pzjpjq4nqx5ycnywekcmycqz0wjp2nqsfwh3q', formt="wif_compressed"), "cdoAa9maGaiLSBj6NiXEz6sFL4HG92N2WLsJPKbJpRGHNaGEA4UC")
        self.assertEqual(w.privkey('tb1qk8pyjanxrtv9039mz3muzjalfuk9r5xuza255z', formt="wif_compressed"), "cdVd1bycgF2FCDDAQjghgc95o92HBeZ8kcrtri316yeujcQTpjK1")
        self.assertEqual(w.privkey('tb1qdy94n2q5qcp0kg7v9yzwe6wvfkhnvyzjncg43j', formt="wif_compressed"), "cZDNjtfqRh2bQSSAa41FUdRtSMYjdPx7xou28pCwGYPrD7NVebpU")
        self.assertEqual(w.privkey('tb1q6xwxcw6m9ga35687tnu5tstmsvmzjwdngss2ql', formt="wif_compressed"), "cXjdt1r1QXj3ES6uLs1cBe3qd8yxaG1fQuM2XaF5xxQBkr1FMkLc")
        self.assertEqual(w.privkey('tb1qgl92qp6haea2ktxe93umzsyhv9z8u32xp2dzm9', formt="wif_compressed"), "cX8wpzyhhRKregpP77mNioDBJ72ZMDHjLHzLJzV6H8jSaP5JETEF")
        self.assertEqual(w.privkey('tb1ql6m8hpkst40p5g3fp6vkc69fysn8w9cph48y5k', formt="wif_compressed"), "cay1UFGRmi4Pm6dXKg6bd9tfXEuzf3Ak3eikcqeVFyUx4EpkNrzL")


    def test_bip39_seed_bip44_standard(self):
        seed_words = 'treat dwarf wealth gasp brass outside high rent blood crowd make initial'
        self.assertEqual(keystore.bip39_is_checksum_valid(seed_words), (True, True))

        w = Bitcoin().wallet(seed_words)

        self.assertTrue(isinstance(w.keystore, keystore.BIP32_KeyStore))

        self.assertEqual(w.keystore.xprv, 'xprv9zGLcNEb3cHUKizLVBz6RYeE9bEZAVPjH2pD1DEzCnPcsemWc3d3xTao8sfhfUmDLMq6e3RcEMEvJG1Et8dvfL8DV4h7mwm9J6AJsW9WXQD')
        self.assertEqual(w.keystore.xpub, 'xpub6DFh1smUsyqmYD4obDX6ngaxhd53Zx7aeFjoobebm7vbkT6f9awJWFuGzBT9FQJEWFBL7UyhMXtYzRcwDuVbcxtv9Ce2W9eMm4KXLdvdbjv')

        self.assertEqual(w.keystore.xtype, 'p2pkh')

        self.assertEqual(w.new_receiving_address(), '16j7Dqk3Z9DdTdBtHcCVLaNQy9MTgywUUo')
        self.assertEqual(w.new_change_address(), '1GG5bVeWgAp5XW7JLCphse14QaC4qiHyWn')

        self.assertEqual(w.receiving_addresses, ['16j7Dqk3Z9DdTdBtHcCVLaNQy9MTgywUUo'])
        self.assertEqual(w.change_addresses, ['1GG5bVeWgAp5XW7JLCphse14QaC4qiHyWn'])

        self.assertEqual(w.new_receiving_addresses(2), ['1K9UsDbEEe2SfKUMTh8B7ZPyYLHeB3f2j7', '1DpbsAw25KyyKwG6oGPCZBPJK7WnL2bKZ9'])
        self.assertEqual(w.new_change_addresses(2), ['15D5w9aZQpcmGfKoQrrVMHRYGdDzyFDkre', '17ABdZyMVWfU2tZEZWUAXPMsD8wTgcbmum'])

        self.assertEqual(w.receiving_address(100), "14TUbjfoT4EKjdyjjUpP3LSBjXMGf87K2Q")

        self.assertEqual(w.privkey('16j7Dqk3Z9DdTdBtHcCVLaNQy9MTgywUUo', formt="wif_compressed"), "L422visYK6FZnmxowLYEof4ty8FvqEUS7YJuNi6Y2hWUqRuDt3qK")
        self.assertEqual(w.privkey('1K9UsDbEEe2SfKUMTh8B7ZPyYLHeB3f2j7', formt="wif_compressed"), "L1HdAazBSe52DxsHv8tXwS1gPAMEXmCXPrUQRTCqd34naS6d28Mk")
        self.assertEqual(w.privkey('1DpbsAw25KyyKwG6oGPCZBPJK7WnL2bKZ9', formt="wif_compressed"), "Kz9fHKgCQdYBgmHSFBMZmgAWUV7zkx6VvUUhS7LuxHaAZXkiG59v")
        self.assertEqual(w.privkey('1GG5bVeWgAp5XW7JLCphse14QaC4qiHyWn', formt="wif_compressed"), "KxzBNXidRYKuwyeZmHPikUsN4eVXP4agnqbcXToQkMb5sBSahsL3")
        self.assertEqual(w.privkey('15D5w9aZQpcmGfKoQrrVMHRYGdDzyFDkre', formt="wif_compressed"), "L59fod5uqjBgouaUMyTdq5Nt8Dho5xeZvxCuZLF8V77zXb7pLSMC")
        self.assertEqual(w.privkey('17ABdZyMVWfU2tZEZWUAXPMsD8wTgcbmum', formt="wif_compressed"), "L4EyTu44oxXpVynitaUQ9Uf7Sw3EoYEGs1GFXaxb6KQATViskWZP")
        self.assertEqual(w.privkey('14TUbjfoT4EKjdyjjUpP3LSBjXMGf87K2Q', formt="wif_compressed"), "L3n4ZyY1YWV8Kyyn8HbRZNgRoyucUJKyFsbdvjNX7vWr5iRfsyoY")

    def test_bip39_seed_bip44_standard_watch(self):
        xpub = 'xpub6DFh1smUsyqmYD4obDX6ngaxhd53Zx7aeFjoobebm7vbkT6f9awJWFuGzBT9FQJEWFBL7UyhMXtYzRcwDuVbcxtv9Ce2W9eMm4KXLdvdbjv'

        w = Bitcoin().watch_wallet(xpub)

        self.assertTrue(isinstance(w.keystore, keystore.BIP32_KeyStore))

        self.assertEqual(w.keystore.xtype, 'p2pkh')
        self.assertTrue(w.is_watching_only)

        self.assertEqual(w.new_receiving_address(), '16j7Dqk3Z9DdTdBtHcCVLaNQy9MTgywUUo')
        self.assertEqual(w.new_change_address(), '1GG5bVeWgAp5XW7JLCphse14QaC4qiHyWn')

        self.assertEqual(w.receiving_addresses, ['16j7Dqk3Z9DdTdBtHcCVLaNQy9MTgywUUo'])
        self.assertEqual(w.change_addresses, ['1GG5bVeWgAp5XW7JLCphse14QaC4qiHyWn'])

        self.assertEqual(w.new_receiving_addresses(2), ['1K9UsDbEEe2SfKUMTh8B7ZPyYLHeB3f2j7', '1DpbsAw25KyyKwG6oGPCZBPJK7WnL2bKZ9'])
        self.assertEqual(w.new_change_addresses(2), ['15D5w9aZQpcmGfKoQrrVMHRYGdDzyFDkre', '17ABdZyMVWfU2tZEZWUAXPMsD8wTgcbmum'])

        self.assertEqual(w.receiving_address(100), "14TUbjfoT4EKjdyjjUpP3LSBjXMGf87K2Q")


    def test_bip39_seed_bip49_p2sh_segwit(self):
        seed_words = 'treat dwarf wealth gasp brass outside high rent blood crowd make initial'
        self.assertEqual(keystore.bip39_is_checksum_valid(seed_words), (True, True))

        w = Bitcoin().p2wpkh_p2sh_wallet(seed_words)

        self.assertTrue(isinstance(w.keystore, keystore.BIP32_KeyStore))

        self.assertEqual(w.keystore.xprv, 'yprvAJEYHeNEPcyBoQYM7sGCxDiNCTX65u4ANgZuSGTrKN5YCC9MP84SBayrgaMyZV7zvkHrr3HVPTK853s2SPk4EttPazBZBmz6QfDkXeE8Zr7')
        self.assertEqual(w.keystore.xpub, 'ypub6XDth9u8DzXV1tcpDtoDKMf6kVMaVMn1juVWEesTshcX4zUVvfNgjPJLXrD9N7AdTLnbHFL64KmBn3SNaTe69iZYbYCqLCCNPZKbLz9niQ4')

        self.assertEqual(w.keystore.xtype, 'p2wpkh-p2sh')

        self.assertEqual(w.new_receiving_address(), '35ohQTdNykjkF1Mn9nAVEFjupyAtsPAK1W')
        self.assertEqual(w.new_change_address(), '3KaBTcviBLEJajTEMstsA2GWjYoPzPK7Y7')

        self.assertEqual(w.receiving_addresses, ['35ohQTdNykjkF1Mn9nAVEFjupyAtsPAK1W'])
        self.assertEqual(w.change_addresses, ['3KaBTcviBLEJajTEMstsA2GWjYoPzPK7Y7'])

        self.assertEqual(w.new_receiving_addresses(2),
                         ['3DitVjWr2LKgPwadXuEd3QXX76BNJXstzj', '3GMNcTdXjJEzkUrv3QaPn7iBCmVoLH3tV8'])
        self.assertEqual(w.new_change_addresses(2),
                         ['3KaBA2yTSn8e1F9LTT731Jr4cZXfVVGD37', '32s2nRytkeABETk2wGs7rez68LjPrNo8u5'])

        self.assertEqual(w.receiving_address(100), "3CSGipmQKSevBFjfQmbqS2n1kedD1FfoXo")
        self.assertEqual(len(w.receiving_addresses), 4)
        self.assertEqual(len(w.change_addresses), 3)

        self.assertEqual(w.privkey('35ohQTdNykjkF1Mn9nAVEFjupyAtsPAK1W', formt="wif_compressed"),
                         "KyUdS5eoaKr7hhXgUnyTk6nyYZGArzSFy49xc32TsuYLAiAtDTzB")
        self.assertEqual(w.privkey('3DitVjWr2LKgPwadXuEd3QXX76BNJXstzj', formt="wif_compressed"),
                         "L5VccEBER99XTkEg1xtpCHaigfjjiH6DzGsJccnyY7eSaSJd85Yv")
        self.assertEqual(w.privkey('3GMNcTdXjJEzkUrv3QaPn7iBCmVoLH3tV8', formt="wif_compressed"),
                         "KwnEbt7APPfwnm378RDX99wTfxqbSY1SYYK6hiAZd1emuA5Q9Tna")
        self.assertEqual(w.privkey('3CSGipmQKSevBFjfQmbqS2n1kedD1FfoXo', formt="wif_compressed"),
                         "KxFZTZwLxiqkBkBqyxZ8CLz2ZEwVawX6jzAtnh5zvVW2YNTuKZnu")
        self.assertEqual(w.privkey('3KaBTcviBLEJajTEMstsA2GWjYoPzPK7Y7', formt="wif_compressed"),
                         "L4N6uquY9mPxc3HxwPVSFYekFpbmqMXwAdbYQNi9awJrJQ4cms8g")
        self.assertEqual(w.privkey('3KaBA2yTSn8e1F9LTT731Jr4cZXfVVGD37', formt="wif_compressed"),
                         "Kzb58wjceypSETV4CHqJr5a1fiWKZ4xhvAAQr4ZGmHwgZdHFuqW3")
        self.assertEqual(w.privkey('32s2nRytkeABETk2wGs7rez68LjPrNo8u5', formt="wif_compressed"),
                         "L5E56T1ZH1gZGVxqJZ5QexhM5KD55TRK5kaTmA9XhigKxwrSJvwD")

    def test_bip39_seed_bip49_p2sh_segwit_watch(self):
        xpub = 'ypub6XDth9u8DzXV1tcpDtoDKMf6kVMaVMn1juVWEesTshcX4zUVvfNgjPJLXrD9N7AdTLnbHFL64KmBn3SNaTe69iZYbYCqLCCNPZKbLz9niQ4'

        w = Bitcoin().watch_p2wpkh_p2sh_wallet(xpub)

        self.assertTrue(isinstance(w.keystore, keystore.BIP32_KeyStore))

        self.assertEqual(w.keystore.xtype, 'p2wpkh-p2sh')
        self.assertTrue(w.is_watching_only)


        self.assertEqual(w.new_receiving_address(), '35ohQTdNykjkF1Mn9nAVEFjupyAtsPAK1W')
        self.assertEqual(w.new_change_address(), '3KaBTcviBLEJajTEMstsA2GWjYoPzPK7Y7')

        self.assertEqual(w.receiving_addresses, ['35ohQTdNykjkF1Mn9nAVEFjupyAtsPAK1W'])
        self.assertEqual(w.change_addresses, ['3KaBTcviBLEJajTEMstsA2GWjYoPzPK7Y7'])

        self.assertEqual(w.new_receiving_addresses(2),
                         ['3DitVjWr2LKgPwadXuEd3QXX76BNJXstzj', '3GMNcTdXjJEzkUrv3QaPn7iBCmVoLH3tV8'])
        self.assertEqual(w.new_change_addresses(2),
                         ['3KaBA2yTSn8e1F9LTT731Jr4cZXfVVGD37', '32s2nRytkeABETk2wGs7rez68LjPrNo8u5'])

        self.assertEqual(w.receiving_address(100), "3CSGipmQKSevBFjfQmbqS2n1kedD1FfoXo")
        self.assertEqual(len(w.receiving_addresses), 4)
        self.assertEqual(len(w.change_addresses), 3)

    def test_bip39_seed_bip49_p2sh_segwit_testnet(self):
        seed_words = 'treat dwarf wealth gasp brass outside high rent blood crowd make initial'
        self.assertEqual(keystore.bip39_is_checksum_valid(seed_words), (True, True))

        w = Bitcoin(testnet=True).p2wpkh_p2sh_wallet(seed_words)

        self.assertTrue(isinstance(w.keystore, keystore.BIP32_KeyStore))

        self.assertEqual(w.keystore.xprv, 'uprv91cL9abSyRBCUSy1raBjwuWYcpB9f21B3KgCtbvhjYpRMv6d1zMmRSTChtv76Kx84tzktSwpfKXGtxDM9rSEVq1m2g1dz2PJHeiiukzTVzB')
        self.assertEqual(w.keystore.xpub, 'upub5EbgZ68LonjVgw3UxbikK3THAr1e4Uj2QYbogzLKHtMQEiRmZXg1yEmgZALF6dFJKtf7t35eE1vcwiUPeujmhSG7pGz5fYAjgD5nTELevk8')

        self.assertEqual(w.keystore.xtype, 'p2wpkh-p2sh')

        self.assertEqual(w.new_receiving_address(), '2NCUmWfaix7dgepJzpHBa3NqHLVTrZbYn5A')
        self.assertEqual(w.new_change_address(), '2NFbJUWbARWN83L18U6JG5HAjp2Lp4A738w')

        self.assertEqual(w.receiving_addresses, ['2NCUmWfaix7dgepJzpHBa3NqHLVTrZbYn5A'])
        self.assertEqual(w.change_addresses, ['2NFbJUWbARWN83L18U6JG5HAjp2Lp4A738w'])

        self.assertEqual(w.new_receiving_addresses(2),
                         ['2N8P2d23LRUrUGhjmFDHJkELchowAWx8oCP', '2Mt48CJ669Mge4n1nDQbhypoSeKAfhMcMME'])
        self.assertEqual(w.new_change_addresses(2),
                         ['2N52CRv9DQN9jP6mqrEJYh2aHSpzdqCDkT2', '2N8AnS2Z83hMarLbkxfxNqgcFyqb8d93jXz'])

        self.assertEqual(w.receiving_address(100), "2NBZCTLMuxJzTT8KYusrAJuKX7uX7Rrn3yb")
        self.assertEqual(len(w.receiving_addresses), 4)
        self.assertEqual(len(w.change_addresses), 3)

        self.assertEqual(w.privkey('2NCUmWfaix7dgepJzpHBa3NqHLVTrZbYn5A', formt="wif_compressed"),
                         "cPtnJoaMd9i9Bu8n25CvzjzjarCcQovkBfY6MpNmP4pAbjzqXVd5")
        self.assertEqual(w.privkey('2N8P2d23LRUrUGhjmFDHJkELchowAWx8oCP', formt="wif_compressed"),
                         "cUZjQvXZ4ytNVtdnVmVHum4afTGP9Z13XaN4TtFeXVrWEyQMskTg")
        self.assertEqual(w.privkey('2Mt48CJ669Mge4n1nDQbhypoSeKAfhMcMME', formt="wif_compressed"),
                         "cVpKjSsujCYy599X9PEsa1t6gzKDM5KSJ1jq5jSPNKff4yaxjAAm")
        self.assertEqual(w.privkey('2NFbJUWbARWN83L18U6JG5HAjp2Lp4A738w', formt="wif_compressed"),
                         "cQzyrNGXvTkbkwKzMoAxUbcHozZtQXfNspMwbQoP9M3cs5S5YCFd")
        self.assertEqual(w.privkey('2NFbJUWbARWN83L18U6JG5HAjp2Lp4A738w', formt="wif_compressed"),
                         "cQzyrNGXvTkbkwKzMoAxUbcHozZtQXfNspMwbQoP9M3cs5S5YCFd")
        self.assertEqual(w.privkey('2N52CRv9DQN9jP6mqrEJYh2aHSpzdqCDkT2', formt="wif_compressed"),
                         "cQkmYkrwMpWo1koPykJ6LbLFtLmGN7P4SJkoFPdGvto3VzMmb7ZK")
        self.assertEqual(w.privkey('2N8AnS2Z83hMarLbkxfxNqgcFyqb8d93jXz', formt="wif_compressed"),
                         "cUyRoiviymD8b4MFkW9NB7tJ8o5FmVtrDGRo7syPT4FKSBDH3ME5")

    def test_bip39_seed_bip84_native_segwit(self):
        seed_words = 'abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon about'
        self.assertEqual(keystore.bip39_is_checksum_valid(seed_words), (True, True))

        w = Bitcoin().p2wpkh_wallet(seed_words)

        self.assertTrue(isinstance(w.keystore, keystore.BIP32_KeyStore))

        self.assertEqual(w.keystore.xprv, 'zprvAdG4iTXWBoARxkkzNpNh8r6Qag3irQB8PzEMkAFeTRXxHpbF9z4QgEvBRmfvqWvGp42t42nvgGpNgYSJA9iefm1yYNZKEm7z6qUWCroSQnE')
        self.assertEqual(w.keystore.xpub, 'zpub6rFR7y4Q2AijBEqTUquhVz398htDFrtymD9xYYfG1m4wAcvPhXNfE3EfH1r1ADqtfSdVCToUG868RvUUkgDKf31mGDtKsAYz2oz2AGutZYs')

        self.assertEqual(w.keystore.xtype, 'p2wpkh')

        self.assertEqual(w.new_receiving_address(), 'bc1qcr8te4kr609gcawutmrza0j4xv80jy8z306fyu')
        self.assertEqual(w.new_change_address(), 'bc1q8c6fshw2dlwun7ekn9qwf37cu2rn755upcp6el')

        self.assertEqual(w.receiving_addresses, ['bc1qcr8te4kr609gcawutmrza0j4xv80jy8z306fyu'])
        self.assertEqual(w.change_addresses, ['bc1q8c6fshw2dlwun7ekn9qwf37cu2rn755upcp6el'])

        self.assertEqual(w.new_receiving_addresses(2),
                         ['bc1qnjg0jd8228aq7egyzacy8cys3knf9xvrerkf9g', 'bc1qp59yckz4ae5c4efgw2s5wfyvrz0ala7rgvuz8z'])
        self.assertEqual(w.new_change_addresses(2),
                         ['bc1qggnasd834t54yulsep6fta8lpjekv4zj6gv5rf', 'bc1qn8alfh45rlsj44pcdt0f2cadtztgnz4gq3h3uf'])

        self.assertEqual(w.receiving_address(100), "bc1q2m7xsl8hf256as3s6e0pvcgz5n5a0de4ey30rl")
        self.assertEqual(len(w.receiving_addresses), 4)
        self.assertEqual(len(w.change_addresses), 3)

        self.assertEqual(w.privkey('bc1qcr8te4kr609gcawutmrza0j4xv80jy8z306fyu', formt="wif_compressed"),
                         "KyZpNDKnfs94vbrwhJneDi77V6jF64PWPF8x5cdJb8ifgg2DUc9d")
        self.assertEqual(w.privkey('bc1qnjg0jd8228aq7egyzacy8cys3knf9xvrerkf9g', formt="wif_compressed"),
                         "Kxpf5b8p3qX56DKEe5NqWbNUP9MnqoRFzZwHRtsFqhzuvUJsYZCy")
        self.assertEqual(w.privkey('bc1qp59yckz4ae5c4efgw2s5wfyvrz0ala7rgvuz8z', formt="wif_compressed"),
                         "L1WWMekCNikUJwrkmxqXWafFu3gmzJ777WpPGgz5Y1o7U11hrDDs")
        self.assertEqual(w.privkey('bc1q8c6fshw2dlwun7ekn9qwf37cu2rn755upcp6el', formt="wif_compressed"),
                         "KxuoxufJL5csa1Wieb2kp29VNdn92Us8CoaUG3aGtPtcF3AzeXvF")
        self.assertEqual(w.privkey('bc1qggnasd834t54yulsep6fta8lpjekv4zj6gv5rf', formt="wif_compressed"),
                         "KyDKM6os4SNpyCN79CGaZF91vVtzmnragXN7A3qAxVvFDws9jBqh")
        self.assertEqual(w.privkey('bc1qn8alfh45rlsj44pcdt0f2cadtztgnz4gq3h3uf', formt="wif_compressed"),
                         "KwQ1irwXPWRmfEH7CtyiFSA2dZ8TCrczAa7onpr3yNEEPrxGmnfS")
        self.assertEqual(w.privkey('bc1q2m7xsl8hf256as3s6e0pvcgz5n5a0de4ey30rl', formt="wif_compressed"),
                         "L2MQjNV27qN6eZy3msVu6KQw8iHzSGMwoNZV7SH9uL5mCnpLcxxw")

    def test_bip39_seed_bip84_native_segwit_watch(self):
        xpub = 'zpub6rFR7y4Q2AijBEqTUquhVz398htDFrtymD9xYYfG1m4wAcvPhXNfE3EfH1r1ADqtfSdVCToUG868RvUUkgDKf31mGDtKsAYz2oz2AGutZYs'

        w = Bitcoin().watch_p2wpkh_wallet(xpub)

        self.assertTrue(isinstance(w.keystore, keystore.BIP32_KeyStore))

        self.assertEqual(w.keystore.xtype, 'p2wpkh')
        self.assertTrue(w.is_watching_only)

        self.assertEqual(w.new_receiving_address(), 'bc1qcr8te4kr609gcawutmrza0j4xv80jy8z306fyu')
        self.assertEqual(w.new_change_address(), 'bc1q8c6fshw2dlwun7ekn9qwf37cu2rn755upcp6el')

        self.assertEqual(w.receiving_addresses, ['bc1qcr8te4kr609gcawutmrza0j4xv80jy8z306fyu'])
        self.assertEqual(w.change_addresses, ['bc1q8c6fshw2dlwun7ekn9qwf37cu2rn755upcp6el'])

        self.assertEqual(w.new_receiving_addresses(2),
                         ['bc1qnjg0jd8228aq7egyzacy8cys3knf9xvrerkf9g', 'bc1qp59yckz4ae5c4efgw2s5wfyvrz0ala7rgvuz8z'])
        self.assertEqual(w.new_change_addresses(2),
                         ['bc1qggnasd834t54yulsep6fta8lpjekv4zj6gv5rf', 'bc1qn8alfh45rlsj44pcdt0f2cadtztgnz4gq3h3uf'])

        self.assertEqual(w.receiving_address(100), "bc1q2m7xsl8hf256as3s6e0pvcgz5n5a0de4ey30rl")
        self.assertEqual(len(w.receiving_addresses), 4)
        self.assertEqual(len(w.change_addresses), 3)

    def test_bip39_seed_bip84_native_segwit_testnet(self):
        seed_words = 'abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon abandon about'
        self.assertEqual(keystore.bip39_is_checksum_valid(seed_words), (True, True))

        w = Bitcoin(testnet=True).p2wpkh_wallet(seed_words)

        self.assertTrue(isinstance(w.keystore, keystore.BIP32_KeyStore))

        self.assertEqual(w.keystore.xprv,
                         'tprv8fSjiqEQ8YG7Ro7gw2ScwcvweYuuWi1ZzGUtrPz918HvDtBzL5s2voFTrN4y3yUwj5cYD54pLhxk6NKCzHUjcka3zbKjbTEcsuAnkzbjhkL')
        self.assertEqual(w.keystore.xpub,
                         'tpubDC8msFGeGuwnKG9Upg7DM2b4DaRqg3CUZa5g8v2SRQ6K4NSkxUgd7HsL2XVWbVm39yBA4LAxysQAm397zwQSQoQgewGiYZqrA9DsP4zbQ1M')

        self.assertEqual(w.keystore.xtype, 'p2wpkh')

        self.assertEqual(w.new_receiving_address(), 'tb1q6rz28mcfaxtmd6v789l9rrlrusdprr9pqcpvkl')
        self.assertEqual(w.new_change_address(), 'tb1q9u62588spffmq4dzjxsr5l297znf3z6j5p2688')

        self.assertEqual(w.receiving_addresses, ['tb1q6rz28mcfaxtmd6v789l9rrlrusdprr9pqcpvkl'])
        self.assertEqual(w.change_addresses, ['tb1q9u62588spffmq4dzjxsr5l297znf3z6j5p2688'])

        self.assertEqual(w.new_receiving_addresses(2),
                         ['tb1qd7spv5q28348xl4myc8zmh983w5jx32cjhkn97', 'tb1qxdyjf6h5d6qxap4n2dap97q4j5ps6ua8sll0ct'])
        self.assertEqual(w.new_change_addresses(2),
                         ['tb1qkwgskuzmmwwvqajnyr7yp9hgvh5y45kg8wvdmd', 'tb1q2vma00td2g9llw8hwa8ny3r774rtt7aenfn5zu'])

        self.assertEqual(w.receiving_address(100), "tb1q0gzpnp4yhr20p7tqdrqmm7872w0unmawzka9zu")
        self.assertEqual(len(w.receiving_addresses), 4)
        self.assertEqual(len(w.change_addresses), 3)

        self.assertEqual(w.privkey('tb1q6rz28mcfaxtmd6v789l9rrlrusdprr9pqcpvkl', formt="wif_compressed"),
                         "cTGhosGriPpuGA586jemcuH9pE9spwUmneMBmYYzrQEbY92DJrbo")
        self.assertEqual(w.privkey('tb1qd7spv5q28348xl4myc8zmh983w5jx32cjhkn97', formt="wif_compressed"),
                         "cQFUndrpAyMaE3HAsjMCXiT94MzfsABCREat1x7Qe3Mtq9KihD4V")
        self.assertEqual(w.privkey('tb1qxdyjf6h5d6qxap4n2dap97q4j5ps6ua8sll0ct', formt="wif_compressed"),
                         "cRe5KDj3rcZJAtZVmWe3G2rdGdXyKCjVbWBVDXqSg2WHq1qq6MNe")
        self.assertEqual(w.privkey('tb1q9u62588spffmq4dzjxsr5l297znf3z6j5p2688', formt="wif_compressed"),
                         "cQ5TcujvBuw5bZfEVJmXZw5Ac1mQ6ryAjHeDrJcmn6oDyhfva5by")
        self.assertEqual(w.privkey('tb1qkwgskuzmmwwvqajnyr7yp9hgvh5y45kg8wvdmd', formt="wif_compressed"),
                         "cVg4cGUL42Kf2XVPC5Uqa7ggsqqWHsKjnZgU6f7nARFoYCVxzh2Y")
        self.assertEqual(w.privkey('tb1q2vma00td2g9llw8hwa8ny3r774rtt7aenfn5zu', formt="wif_compressed"),
                         "cMqRXWvu4JoYcPHMEEJAA48XfmNaf77DH4Tyc9UdbGgKLLEiEqW9")
        self.assertEqual(w.privkey('tb1q0gzpnp4yhr20p7tqdrqmm7872w0unmawzka9zu', formt="wif_compressed"),
                         "cRkgP842XmEK8yr8uDA8QD5kXfBtoTyhjtvGkQTLeXPyZdFWUr2v")


if __name__ == '__main__':
    unittest.main()



from typing import TypedDict, Union, Callable, List, Dict, Any, Literal, Awaitable
from typing_extensions import NotRequired


class ElectrumXBlockCPResponse(TypedDict):
    branch: str
    header: str
    root: str


ElectrumXBlockResponse = Union[str, ElectrumXBlockCPResponse]


class ElectrumXBlockHeadersResponse(TypedDict):
    count: int
    hex: str
    max: int
    root: NotRequired[str]
    branch: NotRequired[str]


class ElectrumXBlockHeaderNotification(TypedDict):
    height: int
    hex: str


BlockHeaderNotificationCallback = Callable[[ElectrumXBlockHeaderNotification], Awaitable[None]]


class ElectrumXBalanceResponse(TypedDict):
    confirmed: int
    unconfirmed: int


class ElectrumXMultiBalanceResponse(TypedDict):
    confirmed: int
    unconfirmed: int
    address: str


class ElectrumXTx(TypedDict):
    height: int
    tx_hash: str
    fee: NotRequired[int]
    tx_pos: NotRequired[int]
    value: NotRequired[int]
    address: NotRequired[str]


ElectrumXHistoryResponse = List[ElectrumXTx]


ElectrumXMempoolResponse = List[ElectrumXTx]


ElectrumXUnspentResponse = List[ElectrumXTx]


class ElectrumXTxAddress(TypedDict):
    height: int
    tx_hash: str
    fee: NotRequired[int]
    tx_pos: NotRequired[int]
    value: NotRequired[int]
    address: str


ElectrumXMultiTxResponse = List[ElectrumXTxAddress]


class ElectrumXScripthashNotification(TypedDict):
    scripthash: str
    status: str


AddressNotificationCallback = Callable[[ElectrumXScripthashNotification], Awaitable[None]]


class ElectrumXVerboseTX(TypedDict):
    blockhash: str
    blocktime: int
    confirmations: int
    hash: str
    hex: str
    locktime: int
    size: int
    time: int
    txid: str
    version: int
    vin: List[Dict[str, Any]]
    vout: List[Dict[str, Any]]
    vsize: int
    weight: int


ElectrumXGetTxResponse = Union[str, ElectrumXVerboseTX]


class ElectrumXMerkleResponse(TypedDict):
    block_height: int
    merkle: List[str]
    pos: int


TxidOrTx = Literal['txid', 'tx']
TargetType = Literal['block_hash', 'block_header', "merkle_root"]


class ElectrumXTSCMerkleResponse(TypedDict):
    composite: bool
    index: int
    nodes: List[str]
    proofType: Literal['branch', 'tree']
    target: str
    targetType: TargetType
    txOrId: str
